/*
								+-----------------------------------+
								|                                   |
								|  ***  LRD classifier engine  ***  |
								|                                   |
								|   Copyright   -tHE SWINe- 2007   |
								|                                   |
								|          Classifier.cpp           |
								|                                   |
								+-----------------------------------+
*/

#include <vector>
#include "../../berLame_src/CallStack.h"
#ifdef WIN32
#include <windows.h> // gl.h needs some typedefs
#endif
#include <GL/gl.h>
#include <GL/glext.h>
#include <math.h>
#include <algorithm>
#include "../../berLame_src/OpenGL20.h"
#include "../../berLame_src/OpenGLState.h"
#include "../../berLame_src/Shader2.h"
#include "../../berLame_src/ShaderSys.h"
#include "../../berLame_src/Texture.h"
#include "../../berLame_src/RenderBuffer2.h"
#include "Load_XML.h"
#include "ConvPlan.h"
#include "RasCalib.h"
#include "Classifier.h"
#include "ScopeGuard.h"

#if defined(_MSC_VER) && !defined(__MWERKS__) && !defined(for)
#define for if(0) {} else for
#endif

extern bool Save_Gray_TGA(const char *p_s_file, int n_width, int n_height,
	const unsigned char *p_buffer);
// debug

/*
 *								=== CGPU_AdaBoost::TClassifierData ===
 */

/*
 *	CGPU_AdaBoost::TClassifierData::TClassifierData(CGLState *p_state,
 *		int n_image_width, int n_image_height, const TWeakClassifier *_p_classifier,
 *		CConvolutionManager &r_conv_plan)
 *		- default constructor
 *		- p_state is OpenGL state guard (needed for rasterizer calibration)
 *		- n_image_width and n_image_height are input image width and height
 *		- _p_classifier contains weak classifier parameters as loaded from xml
 *		- r_conv_plan is convolution plan (contains convolution image neccessary
 *		  to evaluate for this classifier)
 *		- if any error occured, convolution image is set to 0 (p_Convolution() returns 0)
 */
CGPU_AdaBoost::TClassifierData::TClassifierData(CGLState *p_state, int n_image_width,
	int n_image_height, const TWeakClassifier *_p_classifier, CConvolutionManager &r_conv_plan)
	:p_convolution(0), p_classifier(_p_classifier)
{
	for(int i = 0; i < r_conv_plan.n_Convolution_Num(); ++ i) {
		if(r_conv_plan.r_Convolution_Image(i).n_block_width == _p_classifier->n_block_width &&
		   r_conv_plan.r_Convolution_Image(i).n_block_height == _p_classifier->n_block_height) {
			p_convolution = r_conv_plan.r_Convolution_Image(i).p_texture;
			break;
		}
	}
	if(!p_convolution)
		return;
	// finds proper convolution image

	float f_off_s = (float)_p_classifier->n_position_x / n_image_width;
	float f_off_t = (float)_p_classifier->n_position_y / n_image_height;
	CRasterizerCalibrator::TTexCoordRect t_orig_texcoords(f_off_s, f_off_t,
		1 - 1.0f / n_image_width + f_off_s, 1 - 1.0f / n_image_width + f_off_t);
	// texture coordinates

	t_texcoords = t_orig_texcoords;
	if(!CRasterizerCalibrator::CalibrateRasterizer(p_state,
	   n_image_width, n_image_height, t_orig_texcoords, t_texcoords, 1e-6f)) {
		p_convolution = 0; // mark error
		return;
	}
	// calibrate 'em so shader gets what it needs

	int n_slice_num = _p_classifier->n_block_width * _p_classifier->n_block_height;
	v_block_to_slice_x = (float)n_image_width / n_slice_num;
	v_block_to_slice_x_modulo =
		(float)_p_classifier->n_block_width / n_slice_num;
	v_block_to_slice_y = (float)n_image_height / n_slice_num *
		_p_classifier->n_block_width;
	v_block_to_slice_z_offset = .1f / (n_slice_num * (n_image_width / 256));
	// calculate slice selection params for the shader

	v_pixel_x = 1.0f / (n_image_width / _p_classifier->n_block_width);
	v_pixel_y = 1.0f / (n_image_height / _p_classifier->n_block_height);
	// calculate size of 3D texture's pixels

	for(int n_pass = 0; n_pass < 4; ++ n_pass) {
		unsigned char p_mask[4] = {0};
		for(unsigned int i = 0; i < 32; ++ i) {
			int x = (i + _p_classifier->n_position_x) / _p_classifier->n_block_width;
			if(x % 4 == n_pass)
				p_mask[i / 8] |= (unsigned)0x80 >> (i % 8);
		}
		// determine horizontal mask profile

		for(int i = 0; i < 32 * 4; ++ i)
			p_stipple_mask[n_pass][i] = p_mask[i % 4];
		// copy verticaly
	}
	// generate stipple masks
}

/*
 *	bool CGPU_AdaBoost::TClassifierData::operator <(const TClassifierData &r_t_data) const
 *		- less-than operator for weak classifier sorting
 *		- sorts by convolution size, then by block shift
 *		- returns true if this classifier has smaller convolution
 *		  or equal convolution and smaller shift than r_t_data classifier,
 *		  otherwise returns false
 */
bool CGPU_AdaBoost::TClassifierData::operator <(const TClassifierData &r_t_data) const
{
	if(p_classifier->n_block_width == r_t_data.p_classifier->n_block_width &&
	   p_classifier->n_block_height == r_t_data.p_classifier->n_block_height) {
		_ASSERTE(p_convolution == r_t_data.p_convolution);
		// in case block size equals, texture must be identical

		int n_sh0_x = p_classifier->n_position_x % p_classifier->n_block_width;
		int n_sh0_y = p_classifier->n_position_y % p_classifier->n_block_height;
		int n_sh1_x = r_t_data.p_classifier->n_position_x % p_classifier->n_block_width;
		int n_sh1_y = r_t_data.p_classifier->n_position_y % p_classifier->n_block_height;
		// calculate sub-block shift which will affect layer selection

		return n_sh0_x < n_sh1_x || (n_sh0_x == n_sh1_x && n_sh0_y < n_sh1_y);
		// sort by classifier shift (primary key is x, secondary y)
	}

	return p_classifier->n_block_height < r_t_data.p_classifier->n_block_height ||
		(p_classifier->n_block_height == r_t_data.p_classifier->n_block_height &&
		p_classifier->n_block_width < r_t_data.p_classifier->n_block_width);
	// sort by convolution size
}

/*
 *								=== ~CGPU_AdaBoost::TClassifierData ===
 */

/*
 *								=== CGPU_AdaBoost ===
 */

/*
 *	CGPU_AdaBoost::CGPU_AdaBoost(CGLState *p_state, int n_image_width, int n_image_height,
 *		const TAdaBoostClassifier &r_t_classifier, CConvolutionManager &r_conv_plan)
 *		- default constructor
 *		- creates AdaBoost processing plan
 *		- p_state is OpenGL state guard (needed for rasterizer calibration)
 *		- n_image_width and n_image_height are input image width and height
 *		- r_t_classifier contains weak classifiers as loaded from xml
 *		- r_conv_plan is convolution plan (contains convolution images neccessary
 *		  to evaluate for those classifiers)
 *		- if any error occured, classifier list is cleared (number of classifiers returns 0)
 */
CGPU_AdaBoost::CGPU_AdaBoost(CGLState *p_state, int n_image_width, int n_image_height,
	const TAdaBoostClassifier &r_t_classifier, CConvolutionManager &r_conv_plan)
{
	m_classifier_data.reserve(r_t_classifier.n_weak_classifier_num);
	if(m_classifier_data.capacity() < r_t_classifier.n_weak_classifier_num)
		return;
	// alloc classifier list

	for(int i = 0; i < r_t_classifier.n_weak_classifier_num; ++ i) {
		m_classifier_data.push_back(TClassifierData(p_state, n_image_width, n_image_height,
			r_t_classifier.p_weak_classifier + i, r_conv_plan));
		if(!m_classifier_data.back().p_Convolution()) {
			m_classifier_data.clear(); // mark error
			return;
		}
	}
	// make list of classifiers along with coresponding convolution textures

	std::sort(m_classifier_data.begin(), m_classifier_data.end());
	// sort by source convolution image
}

/*
 *	static bool CGPU_AdaBoost::CompileLinkReport(CGLState *p_state, CGLShaderObject &r_shader)
 *		- compiles shader object r_shader and writes any errors to stderr
 *		- p_state is OpenGL state guard
 *		- returns true on success (shader was compiled and linked), false on failure
 */
bool CGPU_AdaBoost::CompileLinkReport(CGLState *p_state, CGLShaderObject &r_shader)
{
	char *p_s_compile = 0, *p_s_link = 0;
	bool b_result = r_shader.Compile_Link(p_state, p_s_compile, p_s_link);
	if(p_s_compile) {
		printf("\n=== %s compiler report ===\n%s\n",
			r_shader.p_FragmentShader_Info()->p_s_name, p_s_compile);
		delete[] p_s_compile;
	}
	if(p_s_link) {
		printf("\n=== %s linker report ===\n%s\n",
			r_shader.p_FragmentShader_Info()->p_s_name, p_s_link);
		delete[] p_s_link;
	}
	return b_result;
}

/*
 *	static int CGPU_AdaBoost::n_AreaSum(const unsigned char *p_image, int n_width,
 *		int n_x0, int n_y0, int n_x1, int n_y1)
 *		- calculates sum of all pixels of greyscale image p_image which is n_width
 *		  pixels wide under rectangular window n_x0 <= x < n_x1, n_y0 <= y < n_y1
 *		- returns pixel sum
 */
int CGPU_AdaBoost::n_AreaSum(const unsigned char *p_image, int n_width,
	int n_x0, int n_y0, int n_x1, int n_y1)
{
	int n_sum = 0;
	for(; n_x0 < n_x1; ++ n_x0) {
		for(int n_y = n_y0; n_y < n_y1; ++ n_y)
			n_sum += p_image[n_x0 + n_y * n_width];
	}
	return n_sum;
}

// todo - document this
bool CGPU_AdaBoost::Validate_ClassifierSampling(CGLState *p_state, CGLTexture_2D *p_source_image,
	CConvolutionManager &r_conv_plan, int &r_n_max_error, float &r_f_std_deviation)
{
	if(!r_conv_plan.Convolve(p_state, p_source_image))
		return false;
	// make sure all convolutions contain correct data

	unsigned char *p_image_buffers;
	int n_image_width = p_source_image->n_Width();
	int n_image_height = p_source_image->n_Height();
	int n_image_size = n_image_width * n_image_height;
	if(!(p_image_buffers = new unsigned char[n_image_size * 6]))
		return false;
	CVectorScopeGuard<unsigned char*> img_del(p_image_buffers);
	// need image buffer for original image (so CPU convolver has a source data),
	// buffer for CPU convolution and for GPU convolution

	float *p_result_buffer;
	if(!(p_result_buffer = new float[n_image_size]))
		return false;
	CVectorScopeGuard<float*> img_del2(p_result_buffer);
	memset(p_result_buffer, 0, n_image_size * sizeof(float));
	// alloc buffer for result

	unsigned char *p_src_buffer = p_image_buffers;
	unsigned char *p_aux_buffer0 = p_image_buffers + n_image_size;
	unsigned char *p_aux_buffer1 = p_image_buffers + n_image_size * 2; // RGBA!
	// pointers to the image buffers

	glPixelStorei(GL_PACK_ALIGNMENT, 1);
	// byte-aligned pixels

	p_source_image->Bind_Enable(p_state);
	glGetTexImage(GL_TEXTURE_2D, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, p_src_buffer);
	// get source image data

	CGLRenderBuffer_FBO *p_render_buffer;
	if(!(p_render_buffer = new CGLRenderBuffer_FBO(n_image_width,
	   n_image_height, 1, true, GL_RGBA8, false, false, 0, false, false, 0)))
		return false;
	CScalarScopeGuard<CGLRenderBuffer_FBO*> rb_del(p_render_buffer);
	if(!p_render_buffer->b_Status() || glGetError() != GL_NO_ERROR)
		return false;
	// create a new render buffer

	CGLTexture_2D *p_result_texture;
	if(!(p_result_texture = new CGLTexture_2D(p_state, n_image_width, n_image_height,
	   GL_RGBA8, false, 0)))
		return false;
	CScalarScopeGuard<CGLTexture_2D*> tex_del(p_result_texture);
	if(!p_result_texture->b_Status())
		return false;
	// create empty texture to render to

	TShaderInfo slice_select_shader("slice_select_shader", true,
		"uniform vec4 v_block_to_slice;\n" // x and y scaling, y is multiplied by block width. z coord contains half-pixel shift
		"void main()\n"
		"{\n"
		"    vec2 v_slice = v_block_to_slice.xy * gl_TexCoord[0].xy;\n"
		"    v_slice.x = mod(v_slice.x, v_block_to_slice.w);\n"
		"    float f_slice = v_slice.x + v_slice.y + v_block_to_slice.z;\n"
		"    f_slice = mod(f_slice, 1.0);\n" // mod is not necessary when sampling texture
		"    gl_FragColor = vec4(f_slice, f_slice, f_slice, 1);\n"
		"}\n" // calculates 3D texture slice
	);
	CGLShaderObject sh_slice_selection(0, &slice_select_shader);
	if(!CompileLinkReport(p_state, sh_slice_selection))
		return false;
	// create slice selection shader

	TShaderInfo test_sample_shader0("test_sample_shader0", true,
		"uniform sampler3D n_texture;\n"
		"uniform vec4 v_block_to_slice;\n" // x and y scaling, y is multiplied by block width. z coord contains half-pixel shift
		"void main()\n"
		"{\n"
		"    vec2 v_slice = v_block_to_slice.xy * gl_TexCoord[0].xy;\n"
		"    v_slice.x = mod(v_slice.x, v_block_to_slice.w);\n"
		"    float f_slice = v_slice.x + v_slice.y + v_block_to_slice.z;\n"
		"    f_slice = mod(f_slice, 1.0);\n" // mod is not necessary when sampling texture
		"    vec3 v_texcoord = vec3(gl_TexCoord[0].xy, f_slice);"
		"    vec3 v_sums0;\n"
		"    v_sums0 = texture3D(n_texture, v_texcoord).xyz;\n"
		"    gl_FragColor = vec4(v_sums0, 1.0);\n"
		"}\n" // calculates 3D texture slice
	);
	CGLShaderObject sh_test_sample0(0, &test_sample_shader0);
	if(!CompileLinkReport(p_state, sh_test_sample0))
		return false;

	TShaderInfo test_sample_shader1("test_sample_shader1", true,
		"uniform sampler3D n_texture;\n"
		"uniform vec4 v_block_to_slice;\n" // x and y scaling, y is multiplied by block width. z coord contains half-pixel shift
		"void main()\n"
		"{\n"
		"    vec2 v_slice = v_block_to_slice.xy * gl_TexCoord[0].xy;\n"
		"    v_slice.x = mod(v_slice.x, v_block_to_slice.w);\n"
		"    float f_slice = v_slice.x + v_slice.y + v_block_to_slice.z;\n"
		"    f_slice = mod(f_slice, 1.0);\n" // mod is not necessary when sampling texture
		"    vec3 v_texcoord = vec3(gl_TexCoord[0].xy, f_slice);"
		"    vec3 v_sums0;\n"
		"    v_sums0 = texture3D(n_texture, v_texcoord).yzw;\n"
		"    gl_FragColor = vec4(v_sums0, 1.0);\n"
		"}\n" // calculates 3D texture slice
	);
	CGLShaderObject sh_test_sample1(0, &test_sample_shader1);
	if(!CompileLinkReport(p_state, sh_test_sample1))
		return false;

	TShaderInfo test_sample_shader2("test_sample_shader2", true,
		"uniform sampler3D n_texture;\n"
		"uniform vec4 v_block_to_slice;\n" // x and y scaling, y is multiplied by block width. z coord contains half-pixel shift
		"uniform vec2 v_pixel;\n" // "large" 3d-tex slice pixel, not viewport pixel
		"void main()\n"
		"{\n"
		"    vec2 v_slice = v_block_to_slice.xy * gl_TexCoord[0].xy;\n"
		"    v_slice.x = mod(v_slice.x, v_block_to_slice.w);\n"
		"    float f_slice = v_slice.x + v_slice.y + v_block_to_slice.z;\n"
		"    f_slice = mod(f_slice, 1.0);\n" // mod is not necessary when sampling texture
		"    vec3 v_texcoord = vec3(gl_TexCoord[0].xy, f_slice);\n"
		"    vec3 v_sums0;\n"
		"    v_sums0.xy = texture3D(n_texture, v_texcoord).zw;\n"
		"    v_texcoord.x += 2.0 * v_pixel.x;\n" // h_ack! intentionally letting there the implicit int to float conversion error
		"    v_sums0.z = texture3D(n_texture, v_texcoord).x;\n"
		"    gl_FragColor = vec4(v_sums0, 1.0);\n"
		"}\n" // calculates 3D texture slice
	);
	CGLShaderObject sh_test_sample2(0, &test_sample_shader2);
	if(!CompileLinkReport(p_state, sh_test_sample2))
		return false;

	TShaderInfo test_sample_shader3("test_sample_shader3", true,
		"uniform sampler3D n_texture;\n"
		"uniform vec4 v_block_to_slice;\n" // x and y scaling, y is multiplied by block width. z coord contains half-pixel shift
		"uniform vec2 v_pixel;\n" // "large" 3d-tex slice pixel, not viewport pixel
		"void main()\n"
		"{\n"
		"    vec2 v_slice = v_block_to_slice.xy * gl_TexCoord[0].xy;\n"
		"    v_slice.x = mod(v_slice.x, v_block_to_slice.w);\n"
		"    float f_slice = v_slice.x + v_slice.y + v_block_to_slice.z;\n"
		"    f_slice = mod(f_slice, 1.0);\n" // mod is not necessary when sampling texture
		"    vec3 v_texcoord = vec3(gl_TexCoord[0].xy, f_slice);"
		"    vec3 v_sums0;\n"
		"    v_sums0.x = texture3D(n_texture, v_texcoord).w;\n"
		"    v_texcoord.x += v_pixel.x;\n"
		"    v_sums0.yz = texture3D(n_texture, v_texcoord).xy;\n"
		"    gl_FragColor = vec4(v_sums0, 1.0);\n"
		"}\n" // calculates 3D texture slice
	);
	CGLShaderObject sh_test_sample3(0, &test_sample_shader3);
	if(!CompileLinkReport(p_state, sh_test_sample3))
		return false;
	// create test sampling shaders

	CGLShaderObject *p_test_sample_sh[4] = {
		&sh_test_sample0, &sh_test_sample1, &sh_test_sample2, &sh_test_sample3
	};

	p_state->BindProgramObject_GL2(0);
	// make sure no shader is active

	int n_max_error = 0;
	double f_sum_diff2 = 0;
	unsigned long n_cmp_num = 0;

	const CGLTexture_3D *p_prev_texture = 0;
	for(int i = 0; i < m_classifier_data.size(); ++ i) {
		TClassifierData &r_t_cl = m_classifier_data[i];
		// get some classifier we're going to work on

		if(p_prev_texture != r_t_cl.p_Convolution()) {
			(p_prev_texture = r_t_cl.p_Convolution())->Bind_Enable(p_state);
			glPixelStorei(GL_PACK_ALIGNMENT, 1);
			glGetTexImage(GL_TEXTURE_3D, 0, GL_RGBA, GL_UNSIGNED_BYTE, p_aux_buffer0);
		}
		// get proper texture (as 3D data, though)

		float f_sum_difference2 = 0;
		unsigned long n_comparison_num = 1;

		const int n_x_cutoff = r_t_cl.n_Offset_X() + r_t_cl.n_Grid_Width() * r_t_cl.n_Block_Width();
		const int n_y_cutoff = r_t_cl.n_Offset_Y() + r_t_cl.n_Grid_Height() * r_t_cl.n_Block_Height();
		for(int y = 0; y < n_image_height - n_x_cutoff; ++ y) {
			for(int x = 0; x < n_image_width - n_y_cutoff; ++ x) {
				// x and y are origin of our image window

				int sums[9];
				{
					const int gridSizeX = r_t_cl.n_Grid_Width();
					const int gridSizeY = r_t_cl.n_Grid_Height();
					int *currentSum = sums;
					const int featurePosX = r_t_cl.n_Offset_X();
					const int featurePosY = r_t_cl.n_Offset_Y();
					const int convolutionSizeX = r_t_cl.n_Block_Width();
					const int convolutionSizeY = r_t_cl.n_Block_Height();
					// feed to michal's pseudocode

					for(int yy = 0; yy < gridSizeY; ++ yy) {
						for(int xx = 0; xx < gridSizeX; ++ xx) {
							*currentSum = n_AreaSum(p_src_buffer + x + y * n_image_width,
								n_image_width, featurePosX + xx * convolutionSizeX,
								featurePosY + yy * convolutionSizeY,
								featurePosX + xx * convolutionSizeX + convolutionSizeX,
								featurePosY + yy * convolutionSizeY + convolutionSizeY);

							*currentSum += convolutionSizeX * convolutionSizeY / 2;
							*currentSum /= convolutionSizeX * convolutionSizeY;
							// change to original algorithm - want result in 0-255 range,
							// also round values to get closer to GPU results

							++ currentSum;
						} 
					}
				}
				// calculate gridSizeX per gridSizeY sums, using Michal's reference algorithm

				int my_sums[9];
				{
					const int n_slice_w = r_t_cl.n_Slice_Width();
					const unsigned char *p_slice_buffer = p_aux_buffer0 +
						n_slice_w * r_t_cl.n_Slice_Height() * r_t_cl.n_Slice(x, y);
					// get proper convolution slice

					int x0 = r_t_cl.n_TexCoord_s(x);
					int y0 = r_t_cl.n_TexCoord_t(y);
					int x1 = x0 + r_t_cl.n_Grid_Width();
					int y1 = y0 + r_t_cl.n_Grid_Height();
					// coords of sum rectangle

					int *p_dest_sum = my_sums;
					for(; y0 < y1; ++ y0) {
						for(int xx = x0; xx < x1; ++ xx) {
							*p_dest_sum = p_slice_buffer[xx + y0 * n_slice_w];
							++ p_dest_sum;
						}
					}
					// get my GPU-calculated convolutions
				}

				const int n_grid_w = r_t_cl.n_Grid_Width();
				const int n_grid_h = r_t_cl.n_Grid_Height();
				for(int j = 0; j < n_grid_w * n_grid_h; ++ j) {
					int n_error = abs(sums[j] != my_sums[j]);
					f_sum_difference2 += n_error * n_error;
					if(n_max_error < n_error)
						n_max_error = n_error;
				}
				n_comparison_num += n_grid_w * n_grid_h;
				// compare results
			}
		}
		// check if convolutions are correct

		for(int y = 0; y < n_image_height - n_x_cutoff; ++ y) {
			for(int x = 0; x < n_image_width - n_y_cutoff; ++ x) {
				// ------------------------------------ shader simulation begin -----

				int n_slice_w = r_t_cl.n_Slice_Width();
				const unsigned int *p_slice_buffer = (const unsigned int*)(p_aux_buffer0 +
					n_slice_w * r_t_cl.n_Slice_Height() * r_t_cl.n_Slice(x, y));
				n_slice_w /= 4;
				// get proper convolution slice (in packed format)

				int s = r_t_cl.n_TexCoord_s(x);
				int t = r_t_cl.n_TexCoord_t(y);
				// texcoords

				unsigned char my_sums[9];
				unsigned int n_sam0, n_sam1, n_sam2;
				if(s % 4 == 0) {
					p_slice_buffer += t * n_slice_w; // cur scanline
					n_sam0 = p_slice_buffer[s / 4];
					p_slice_buffer += n_slice_w;
					n_sam1 = p_slice_buffer[s / 4];
					p_slice_buffer += n_slice_w;
					n_sam2 = p_slice_buffer[s / 4];
				} else if(s % 4 == 1) {
					p_slice_buffer += t * n_slice_w; // cur scanline
					n_sam0 = p_slice_buffer[s / 4] >> 8;
					p_slice_buffer += n_slice_w;
					n_sam1 = p_slice_buffer[s / 4] >> 8;
					p_slice_buffer += n_slice_w;
					n_sam2 = p_slice_buffer[s / 4] >> 8;
				} else if(s % 4 == 2) {
					p_slice_buffer += t * n_slice_w; // cur scanline
					n_sam0 = (p_slice_buffer[s / 4] >> 16) | (p_slice_buffer[s / 4 + 1] << 16);
					p_slice_buffer += n_slice_w;
					n_sam1 = (p_slice_buffer[s / 4] >> 16) | (p_slice_buffer[s / 4 + 1] << 16);
					p_slice_buffer += n_slice_w;
					n_sam2 = (p_slice_buffer[s / 4] >> 16) | (p_slice_buffer[s / 4 + 1] << 16);
				} else /*if(s % 4 == 3)*/ {
					p_slice_buffer += t * n_slice_w; // cur scanline
					n_sam0 = (p_slice_buffer[s / 4] >> 24) | (p_slice_buffer[s / 4 + 1] << 8);
					p_slice_buffer += n_slice_w;
					n_sam1 = (p_slice_buffer[s / 4] >> 24) | (p_slice_buffer[s / 4 + 1] << 8);
					p_slice_buffer += n_slice_w;
					n_sam2 = (p_slice_buffer[s / 4] >> 24) | (p_slice_buffer[s / 4 + 1] << 8);
				}
				// get three GPU-vector-like samples

				my_sums[0] = n_sam0;
				my_sums[1] = n_sam0 >> 8;
				my_sums[2] = n_sam0 >> 16;
				my_sums[3] = n_sam1;
				my_sums[4] = n_sam1 >> 8;
				my_sums[5] = n_sam1 >> 16;
				my_sums[6] = n_sam2;
				my_sums[7] = n_sam2 >> 8;
				my_sums[8] = n_sam2 >> 16;
				// de-swizzle samples

				int n_rank_diff = r_t_cl.n_Alpha_Num() / 2;
				// fixme - classifier array offset

				for(int i = 0, n = r_t_cl.n_Grid_Width() * r_t_cl.n_Grid_Height(); i < n; ++ i) {
					n_rank_diff += my_sums[r_t_cl.n_Block_A()] > my_sums[i];
					n_rank_diff -= my_sums[r_t_cl.n_Block_B()] > my_sums[i];
				}
				// vypocet ranku (todo - ranky se od sebe muzou primo odecitat, netreba
				// dvou promennych)

				p_result_buffer[x + y * n_image_width] += r_t_cl.f_Alpha(n_rank_diff);

				// ------------------------------------ shader simulation end -------
			}
		}
		// simulate shader execution

		f_sum_diff2 += f_sum_difference2;
		n_cmp_num += n_comparison_num;
	}

	int n_slice_mismatch_num = 0, n_slice_test_num = 0;
	for(int i = 0; i < m_classifier_data.size(); ++ i) {
		TClassifierData &r_t_cl = m_classifier_data[i];
		// get some classifier we're going to work on

		p_state->DisableTexture2D();
		p_state->DisableTexture3D();

		int p_viewport[4];
		glGetIntegerv(GL_VIEWPORT, p_viewport);
		p_render_buffer->Bind();
		p_render_buffer->BindColorTexture_2D(*p_result_texture, GL_TEXTURE_2D);
		glViewport(0, 0, p_render_buffer->n_Width(), p_render_buffer->n_Height());
		// bind render buffer

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		// reset matrices

		sh_slice_selection.t_Find_Parameter("v_block_to_slice").SetValue(
			r_t_cl.v_block_to_slice_x, r_t_cl.v_block_to_slice_y,
			r_t_cl.v_block_to_slice_z_offset, r_t_cl.v_block_to_slice_x_modulo);
		sh_slice_selection.Bind(p_state);
		// bind slice selection shader (outputs 3d texture slice for each pixel)

		glBegin(GL_QUADS);
		glTexCoord2f(r_t_cl.t_texcoords.f_min_s, r_t_cl.t_texcoords.f_min_t);
		glVertex2f(-1, -1);
		glTexCoord2f(r_t_cl.t_texcoords.f_max_s, r_t_cl.t_texcoords.f_min_t);
		glVertex2f( 1, -1);
		glTexCoord2f(r_t_cl.t_texcoords.f_max_s, r_t_cl.t_texcoords.f_max_t);
		glVertex2f( 1,  1);
		glTexCoord2f(r_t_cl.t_texcoords.f_min_s, r_t_cl.t_texcoords.f_max_t);
		glVertex2f(-1,  1);
		glEnd();
		// render fullscreen quad

		p_state->BindProgramObject_GL2(0);
		// release shader

		p_render_buffer->BindColorTexture_2D(0, GL_TEXTURE_2D);
		p_render_buffer->Release();
		glViewport(p_viewport[0], p_viewport[1], p_viewport[2], p_viewport[3]);
		// release render-buffer, put back original viewport

		p_result_texture->Bind_Enable(p_state);
		glPixelStorei(GL_PACK_ALIGNMENT, 1);
		glGetTexImage(GL_TEXTURE_2D, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, p_aux_buffer1);

		if(glGetError() != GL_NO_ERROR)
			return false;

		const int n_slice_num = r_t_cl.n_Slice_Num();
		for(int y = 0; y < n_image_height; ++ y) {
			for(int x = 0; x < n_image_width; ++ x) {
				int n_ref_slice = r_t_cl.n_Slice(x, y);
				// calculate image slice no.

				int n_slice = (p_aux_buffer1[x + y * n_image_width] * n_slice_num) / 255;
				// read slice no. generated by shader

				if(n_slice != n_ref_slice)
					++ n_slice_mismatch_num;
				++ n_slice_test_num;
			}
		}
	}
	// determine wheter correct slice is chosen in shader

	if(n_slice_mismatch_num && (float)n_slice_mismatch_num / n_slice_test_num > .02f) {
		fprintf(stderr, "\terror: %f convolution slice mismatch rate\n",
			(float)n_slice_mismatch_num / n_slice_test_num);
		return false;
	} else if(n_slice_mismatch_num) {
		fprintf(stderr, "\twarning: %f convolution slice mismatch rate (tolerable)\n",
			(float)n_slice_mismatch_num / n_slice_test_num);
	} else
		printf("\tconvolution slice matching is well within limits\n");
	// :)

	bool b_feature_sampler_errors = false;
	p_prev_texture = 0;
	for(int i = 0; i < m_classifier_data.size(); ++ i) {
		TClassifierData &r_t_cl = m_classifier_data[i];
		// get some classifier we're going to work on

		p_state->DisableTexture2D();
		r_t_cl.p_Convolution()->Bind_Enable(p_state);
		p_state->Color3f(1, 1, 1);

		int p_viewport[4];
		glGetIntegerv(GL_VIEWPORT, p_viewport);
		p_render_buffer->Bind();
		p_render_buffer->BindColorTexture_2D(*p_result_texture, GL_TEXTURE_2D);
		glViewport(0, 0, p_render_buffer->n_Width(), p_render_buffer->n_Height());
		// bind render buffer

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		// reset matrices

		if(glGetError() != GL_NO_ERROR)
			return false;

		p_state->Enable(GL_POLYGON_STIPPLE);
		// enable stipple mask

		glClear(GL_COLOR_BUFFER_BIT);
		// shouldn't be neccessary ... it's here just for debugging

		for(int n_pass = 0; n_pass < 4; ++ n_pass) {
			glPolygonStipple(r_t_cl.p_Stipple_Mask(n_pass));
			// t_odo - need different stipple/pass patterns for different x-kernel-sizes, investigate into it,
			// design and conquer :) maybe save reference image as well so you don't fix what's not broken

			p_test_sample_sh[n_pass]->t_Find_Parameter("v_block_to_slice").SetValue(
				r_t_cl.v_block_to_slice_x, r_t_cl.v_block_to_slice_y,
				r_t_cl.v_block_to_slice_z_offset, r_t_cl.v_block_to_slice_x_modulo);
			if(n_pass > 1) {
				p_test_sample_sh[n_pass]->t_Find_Parameter("v_pixel").SetValue(r_t_cl.v_pixel_x,
					r_t_cl.v_pixel_y);
			}
			p_test_sample_sh[n_pass]->Bind(p_state);

			glBegin(GL_QUADS);
			glTexCoord2f(r_t_cl.t_texcoords.f_min_s, r_t_cl.t_texcoords.f_min_t);
			glVertex2f(-1, -1);
			glTexCoord2f(r_t_cl.t_texcoords.f_max_s, r_t_cl.t_texcoords.f_min_t);
			glVertex2f( 1, -1);
			glTexCoord2f(r_t_cl.t_texcoords.f_max_s, r_t_cl.t_texcoords.f_max_t);
			glVertex2f( 1,  1);
			glTexCoord2f(r_t_cl.t_texcoords.f_min_s, r_t_cl.t_texcoords.f_max_t);
			glVertex2f(-1,  1);
			glEnd();
			// render fullscreen quad
		}
		// render image in four stippled passes

		p_state->Disable(GL_POLYGON_STIPPLE);
		// disable stipple mask

		p_state->BindProgramObject_GL2(0);
		// release shader

		if(glGetError() != GL_NO_ERROR)
			return false;

		p_render_buffer->BindColorTexture_2D(0, GL_TEXTURE_2D);
		p_render_buffer->Release();
		glViewport(p_viewport[0], p_viewport[1], p_viewport[2], p_viewport[3]);
		// release render-buffer, put back original viewport

		p_result_texture->Bind_Enable(p_state);
		glPixelStorei(GL_PACK_ALIGNMENT, 1);
		glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_UNSIGNED_BYTE, p_aux_buffer1);
		// download image's red components (all first grid samples)

		if(glGetError() != GL_NO_ERROR)
			return false;

		if(p_prev_texture != r_t_cl.p_Convolution()) {
			(p_prev_texture = r_t_cl.p_Convolution())->Bind_Enable(p_state);
			glPixelStorei(GL_PACK_ALIGNMENT, 1);
			glGetTexImage(GL_TEXTURE_3D, 0, GL_RGBA, GL_UNSIGNED_BYTE, p_aux_buffer0);
		}
		// get proper texture (as 3D data, though)

		int n_max_error = 0, n_sample_num = 0, p_error_num[3] = {0};
		float f_deviation = 0;

		const int n_x_cutoff = r_t_cl.n_Offset_X() + r_t_cl.n_Grid_Width() * r_t_cl.n_Block_Width();
		const int n_y_cutoff = r_t_cl.n_Offset_Y() + r_t_cl.n_Grid_Height() * r_t_cl.n_Block_Height();
		for(int y = 0; y < n_image_height - n_y_cutoff; ++ y) {
			for(int x = 0; x < n_image_width - n_x_cutoff; ++ x) {
				// x and y are origin of our image window

				int n_slice_w = r_t_cl.n_Slice_Width();
				const unsigned char *p_slice_buffer = p_aux_buffer0 +
					n_slice_w * r_t_cl.n_Slice_Height() * r_t_cl.n_Slice(x, y);

				int x0 = r_t_cl.n_TexCoord_s(x);
				int y0 = r_t_cl.n_TexCoord_t(y);
				int x1 = x0 + r_t_cl.n_Grid_Width();
				int y1 = y0 + r_t_cl.n_Grid_Height();
				// coords of sum rectangle

				int my_sums[9];
				int *p_dest_sum = my_sums;
				for(; y0 < y1; ++ y0) {
					for(int xx = x0; xx < x1; ++ xx) {
						*p_dest_sum = p_slice_buffer[xx + y0 * n_slice_w];
						++ p_dest_sum;
					}
				}
				// get my GPU-assisted sums

				for(int n = 0; n < 3; ++ n) {
					if(p_aux_buffer1[(x + y * n_image_width) * 4 + n] != my_sums[n]) {
						int n_err = abs(p_aux_buffer1[(x + y * n_image_width) * 4 + n] - my_sums[n]);
						if(n_max_error < n_err)
							n_max_error = n_err;
						f_deviation += n_err * n_err;
						++ p_error_num[n];
					}
				}
				// test all three grid components

				n_sample_num += 3;
			}
		}

		f_deviation /= n_sample_num;
		// standard deviation

		if(n_max_error) {
			printf("\twarning: feature sampling %d has std deviation %f (max error %d) (%d %d %d)\n",
				i, f_deviation, n_max_error, p_error_num[0], p_error_num[1], p_error_num[2]);
			b_feature_sampler_errors = true;
			Save_Gray_TGA("first_components.tga", n_image_width * 4, n_image_height, p_aux_buffer1);
		}
	}
	// test feature sampling/swizzling system

	if(!b_feature_sampler_errors)
		printf("\tfeature grid sampling is well within limits\n");
	else {
		printf("\tfeature grid sampling exhibits some errors\n");
		//return false;
	}
	// feature sampling test

		/*int n_sample_x = (x + r_t_cl.p_classifier->n_position_x) / p_classifier->n_block_width;
		int n_sample_y = (y + r_t_cl.p_classifier->n_position_y) / p_classifier->n_block_height;

		TShaderInfo weakcl_shader("weak_classifier_shader", true,
			"uniform sampler3D n_texture;\n"
			"uniform vec2 v_pixel;\n" // pixel-w, pixel-h
			"uniform float f_block_w;\n" // block width divided by block count
			"\n"
			"void main()\n"
			"{\n"
			"    float f_slice = gl_TexCoord[0].x + gl_TexCoord[0].y * f_block_w;\n"
			// calculate 3D texture slice

			"    vec3 sums0, sums1, sums2;\n"
			// for 3x3 grid

			"    vec3 v_texcoord(gl_TexCoord[0].xy, f_slice);"
			"    sums0 = texture3D(n_texture, v_texcoord);\n"
			"    v_texcoord.y += v_pixel.y;\n"
			"    sums1 = texture3D(n_texture, v_texcoord);\n"
			"    v_texcoord.y += v_pixel.y;\n"
			"    sums2 = texture3D(n_texture, v_texcoord);\n"
			// load block; this code will need some swizzling and sometimes twice as much samples

			"    vec3 blockA(sums0.x);\n"
			"    vec3 blockB(sums1.z);\n"
			// these might end up somewhere else

			"    float f_rank = dot(vec3(greaterThan(blockA, sums0)), vec3(1)) +\n"
			"                   dot(vec3(greaterThan(blockA, sums1)), vec3(1)) +\n"
			"                   dot(vec3(greaterThan(blockA, sums2)), vec3(1)) -;\n"
			"                   dot(vec3(greaterThan(blockB, sums0)), vec3(1)) -\n"
			"                   dot(vec3(greaterThan(blockB, sums1)), vec3(1)) -\n"
			"                   dot(vec3(greaterThan(blockB, sums2)), vec3(1));\n"
			// calculate counts; blockA and blockB might be vec4, there might be more dots

			"    gl_FragColor = vec4(f_rank, f_rank, f_rank, 1);\n"
			// write comparison results
			"}\n");*/

	float *p_src = p_result_buffer;
	unsigned char *p_dest = p_aux_buffer0;
	const int n_window = 24;
	for(int y = 0; y < n_image_height - n_window; ++ y) {
		for(int x = 0; x < n_image_width - n_window; ++ x, ++ p_src, ++ p_dest) {
			*p_dest = (unsigned char)(*p_src < 0)? 0 : ((*p_src >= 1)? 255 : *p_src * 255);
			if(*p_src > 10) {
				for(int xx = 0; xx < n_window; ++ xx) {
					p_src_buffer[x + xx + y * n_image_width] = 0xff;
					p_src_buffer[x + xx + (y + n_window - 1) * n_image_width] = 0xff;
					p_src_buffer[x + (xx + y) * n_image_width] = 0xff;
					p_src_buffer[x + n_window - 1 + (xx + y) * n_image_width] = 0xff;
				}
				// inscribe window to source image
			}
		}
		p_dest += n_window;
		p_src += n_window;
	}
	// convert classification result to ubyte range

	if(!Save_Gray_TGA("classification_result1.tga",
	   n_image_width, n_image_height, p_src_buffer) ||
	   !Save_Gray_TGA("classification_result0.tga",
	   n_image_width, n_image_height, p_aux_buffer0))
		fprintf(stderr, "warning: error saving reference classification results\n");
	// save classifications

	r_f_std_deviation = (float)sqrt(f_sum_diff2 / n_cmp_num);
	r_n_max_error = n_max_error;

	// cleanup is done automatically :)

	return true;
}

/*
 *								=== ~CGPU_AdaBoost ===
 */
