/*
								+-----------------------------------+
								|                                   |
								|  ***  LRD classifier engine  ***  |
								|                                   |
								|   Copyright   -tHE SWINe- 2007   |
								|                                   |
								|           Classifier.h            |
								|                                   |
								+-----------------------------------+
*/

#ifndef __LRD_CLASSIFIER_GPU_ENGINE_INCLUDED
#define __LRD_CLASSIFIER_GPU_ENGINE_INCLUDED

/*

  group classifiers by source convolutions

  create vertex shader that would calculate proper source texcoords for classifiers

  create fragment shader that would group 


-------------------------------------------------------

  classifier results for different pixel offsets come out in different textures
  need to interleave the values; do it:

  using branching (determine pixel offset, use branch that will calculate proper result) - slow

  fill 4 results at once (RGBA pack) - unable to use early alpha test for cutoff, max convolution size will be 4 (4 offsets)

  render to multiple buffers at once - have to interleave results later which could be problematic, requires extra pass,
	limited to 4 texture units anyway

  decision: fill 4 results at once; in case convolutions are larger, use branching

  !!!!! need to use branching anyway, RGBA pack helps with horizontal offsets only

  have different offsets of convolution in different layers of 3d texture

  it's terrible mess, all this ... but on the other hand texture accesses necessary to read 3x3 block are minimized

*/

/*
 *	class CGPU_AdaBoost
 *		- GPU AdaBoost engine class (creates operation plan
 *		  (once) and then executes it (many times))
 */
class CGPU_AdaBoost {
protected:
	/*
	 *	struct CGPU_AdaBoost::TClassifierData
	 *		- data for weak classifier execution
	 *		  (source texture, shader parameters, stipple mask)
	 */
	struct TClassifierData {
	protected:
		const TWeakClassifier *p_classifier;
		// classifier parameters from xml file

		const CGLTexture_3D *p_convolution;
		// convolution image (owned by CConvolutionManager)

		unsigned char p_stipple_mask[4][32 * 4];
		// stipple masks

	public:
		CRasterizerCalibrator::TTexCoordRect t_texcoords;
		// calibrated texcoords for shader

		float v_block_to_slice_x;
		float v_block_to_slice_x_modulo;
		float v_block_to_slice_y;
		float v_block_to_slice_z_offset;
		// slice selection shader parameters

		float v_pixel_x, v_pixel_y;
		// 3d texture texel dimensions

		/*
		 *	CGPU_AdaBoost::TClassifierData::TClassifierData(CGLState *p_state,
		 *		int n_image_width, int n_image_height, const TWeakClassifier *_p_classifier,
		 *		CConvolutionManager &r_conv_plan)
		 *		- default constructor
		 *		- p_state is OpenGL state guard (needed for rasterizer calibration)
		 *		- n_image_width and n_image_height are input image width and height
		 *		- _p_classifier contains weak classifier parameters as loaded from xml
		 *		- r_conv_plan is convolution plan (contains convolution image neccessary
		 *		  to evaluate for this classifier)
		 *		- if any error occured, convolution image is set to 0 (p_Convolution() returns 0)
		 */
		TClassifierData(CGLState *p_state, int n_image_width, int n_image_height,
			const TWeakClassifier *_p_classifier, CConvolutionManager &r_conv_plan);

		/*
		 *	bool CGPU_AdaBoost::TClassifierData::operator <(const TClassifierData &r_t_data) const
		 *		- less-than operator for weak classifier sorting
		 *		- sorts by convolution size, then by block shift
		 *		- returns true if this classifier has smaller convolution
		 *		  or equal convolution and smaller shift than r_t_data classifier,
		 *		  otherwise returns false
		 */
		bool operator <(const TClassifierData &r_t_data) const;

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Slice_Num() const
		 *		- returns number of convlolution texture slices (each one contains convolution
		 *		  with different image shift), also equals to convolution size (kernel area)
		 */
		inline int n_Slice_Num() const
		{
			return p_classifier->n_block_width * p_classifier->n_block_height;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Block_Width() const
		 *		- returns width of convolution kernel
		 */
		inline int n_Block_Width() const
		{
			return p_classifier->n_block_width;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Block_Height() const
		 *		- returns height of convolution kernel
		 */
		inline int n_Block_Height() const
		{
			return p_classifier->n_block_height;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Slice_Width() const
		 *		- returns width of convolved image in samples.
		 *		  that's four times more width in pixels (RGBA packing)
		 */
		inline int n_Slice_Width() const
		{
			return p_convolution->n_Width() * 4;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Slice_Height() const
		 *		- returns height of convolved image (in samples or pixels, no packing here)
		 */
		inline int n_Slice_Height() const
		{
			return p_convolution->n_Height();
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Grid_Width() const
		 *		- returns feature grid width (number of feature horizontal convolutions)
		 */
		inline int n_Grid_Width() const
		{
			return p_classifier->n_width;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Grid_Height() const
		 *		- returns feature grid height (number of feature vertical convolutions)
		 */
		inline int n_Grid_Height() const
		{
			return p_classifier->n_height;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Offset_X() const
		 *		- returns horizontal offset of first sample from window origin in pixels
		 */
		inline int n_Offset_X() const
		{
			return p_classifier->n_position_x;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Offset_Y() const
		 *		- returns vertical offset of first sample from window origin in pixels
		 */
		inline int n_Offset_Y() const
		{
			return p_classifier->n_position_y;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Alpha_Num() const
		 *		- returns number of alphas
		 */
		inline int n_Alpha_Num() const
		{
			return p_classifier->n_alpha_num;
		}

		/*
		 *	inline float CGPU_AdaBoost::TClassifierData::f_Alpha(int n_index) const
		 *		- returns alpha coefficient with zero-based index n_index
		 */
		inline float f_Alpha(int n_index) const
		{
			return p_classifier->p_alpha[n_index];
		}

		/*
		 *	inline const unsigned char *CGPU_AdaBoost::TClassifierData::p_Stipple_Mask(
		 *		int n_pass) const
		 *		- returns stipple mask for pass n_pass (0 to 3)
		 */
		inline const unsigned char *p_Stipple_Mask(int n_pass) const
		{
			return p_stipple_mask[n_pass];
		}

		/*
		 *	const CGLTexture_3D *p_Convolution() const
		 *		- returns convolution texture for this classifier or 0 on error
		 */
		const CGLTexture_3D *p_Convolution() const
		{
			return p_convolution;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Block_A() const
		 *		- returns zero-based index of the first feature block to be compared with others
		 */
		inline int n_Block_A() const
		{
			return p_classifier->n_block_a;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Block_B() const
		 *		- returns zero-based index of second feature block to be compared with others
		 */
		inline int n_Block_B() const
		{
			return p_classifier->n_block_b;
		}

		// todo - add custom fuckin' designed shader for each classifier ... save merging for later

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_Slice(int x, int y) const
		 *		- returns zero-based index of texture slice for x and y in output image
		 */
		inline int n_Slice(int x, int y) const
		{
			return (x + p_classifier->n_position_x) % p_classifier->n_block_width +
				((y + p_classifier->n_position_y) % p_classifier->n_block_height) *
				p_classifier->n_block_width;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_TexCoord_s(int x)
		 *		- returns texture coordinate s for x position in output image (in pixels)
		 */
		inline int n_TexCoord_s(int x)
		{
			return (x + p_classifier->n_position_x) / p_classifier->n_block_width;
		}

		/*
		 *	inline int CGPU_AdaBoost::TClassifierData::n_TexCoord_t(int y)
		 *		- returns texture coordinate t for y position in output image (in pixels)
		 */
		inline int n_TexCoord_t(int y)
		{
			return (y + p_classifier->n_position_y) / p_classifier->n_block_height;
		}
	};

	std::vector<TClassifierData> m_classifier_data;

public:
	/*
	 *	CGPU_AdaBoost::CGPU_AdaBoost(CGLState *p_state, int n_image_width, int n_image_height,
	 *		const TAdaBoostClassifier &r_t_classifier, CConvolutionManager &r_conv_plan)
	 *		- default constructor
	 *		- creates AdaBoost processing plan
	 *		- p_state is OpenGL state guard (needed for rasterizer calibration)
	 *		- n_image_width and n_image_height are input image width and height
	 *		- r_t_classifier contains weak classifiers as loaded from xml
	 *		- r_conv_plan is convolution plan (contains convolution images neccessary
	 *		  to evaluate for those classifiers)
	 *		- if any error occured, classifier list is cleared (number of classifiers returns 0)
	 */
	CGPU_AdaBoost(CGLState *p_state, int n_image_width, int n_image_height,
		const TAdaBoostClassifier &r_t_classifier, CConvolutionManager &r_conv_plan);

	// todo - document this
	bool Validate_ClassifierSampling(CGLState *p_state, CGLTexture_2D *p_source_image,
		CConvolutionManager &r_conv_plan, int &r_n_max_error, float &r_f_std_deviation);

protected:
	static bool CompileLinkReport(CGLState *p_state, CGLShaderObject &r_shader);
	static int n_AreaSum(const unsigned char *p_image, int n_width,
		int n_x0, int n_y0, int n_x1, int n_y1);
};

#endif //__LRD_CLASSIFIER_GPU_ENGINE_INCLUDED
