/*
								+-----------------------------------+
								|                                   |
								|  ***  Convolution scheduler  ***  |
								|                                   |
								|   Copyright   -tHE SWINe- 2007   |
								|                                   |
								|           ConvPlan.cpp            |
								|                                   |
								+-----------------------------------+
*/

#include <vector>
#include "../../berLame_src/CallStack.h"
#include <stdlib.h>
#include <stdio.h>
#include <algorithm>
#include <functional>
#include <math.h>
#ifdef WIN32
#include <windows.h> // gl.h needs some typedefs
#endif
#include <GL/gl.h>
#include <GL/glext.h>
#include "../../berLame_src/OpenGL20.h"
#include "../../berLame_src/OpenGLState.h"
#include "../../berLame_src/VertexBufferObject.h"
#include "../../berLame_src/Shader2.h"
#include "../../berLame_src/ShaderSys.h"
#include "../../berLame_src/Texture.h"
#include "../../berLame_src/RenderBuffer2.h"
#include "Load_XML.h"
#include "ConvPlan.h"
#include "RasCalib.h"

#if defined(_MSC_VER) && !defined(__MWERKS__) && !defined(for)
#define for if(0) {} else for
#endif

/*
 *								=== TConvolutionSize ===
 */

/*
 *	TConvolutionSize::TConvolutionSize()
 *		- default constructor
 */
TConvolutionSize::TConvolutionSize()
{}

/*
 *	TConvolutionSize::TConvolutionSize(const TConvolutionSize &r_t_info)
 *		- copy constructor
 */
TConvolutionSize::TConvolutionSize(const TConvolutionSize &r_t_info)
	:n_block_width(r_t_info.n_block_width), n_block_height(r_t_info.n_block_height)
{}

/*
 *	TConvolutionSize::TConvolutionSize(const TWeakClassifier &r_t_weak_cl)
 *		- constructor creating convolution info satisfying weak classifier demands
 */
TConvolutionSize::TConvolutionSize(const TWeakClassifier &r_t_weak_cl)
	:n_block_width(r_t_weak_cl.n_block_width), n_block_height(r_t_weak_cl.n_block_height)
{}

/*
 *	TConvolutionSize TConvolutionSize::operator =(const TConvolutionSize &r_t_info)
 *		- copy operator
 */
TConvolutionSize TConvolutionSize::operator =(const TConvolutionSize &r_t_info)
{
	n_block_width = r_t_info.n_block_width;
	n_block_height = r_t_info.n_block_height;
	return *this;
}

/*
 *	bool TConvolutionSize::operator ==(const TConvolutionSize &r_t_info)
 *		- equality operator
 */
bool TConvolutionSize::operator ==(const TConvolutionSize &r_t_info)
{
	return n_block_width == r_t_info.n_block_width &&
		n_block_height == r_t_info.n_block_height;
}

/*
 *	bool TConvolutionSize::operator <(const TConvolutionSize &r_t_info)
 *		- less-then operator for sorting
 */
bool TConvolutionSize::operator <(const TConvolutionSize &r_t_info)
{
	return n_block_width < r_t_info.n_block_width ||
		(n_block_width == r_t_info.n_block_width &&
		n_block_height < r_t_info.n_block_height);
}

/*
 *	int TConvolutionSize::Dump()
 *		- prints convolution info dump to stdout
 *		- MSVC can't handle return void; in templates so it returns number of prited chars
 */
int TConvolutionSize::Dump()
{
	return printf("\tconvolution(block-width: %d, block-height: %d)\n",
		n_block_width, n_block_height);
}

/*
 *								=== ~TConvolutionSize ===
 */

/*
 *								=== TConvolutionList ===
 */

/*
 *	void TConvolutionList::Dump()
 *		- prints convolution info dumps to stdout
 */
void TConvolutionList::Dump()
{
	printf("=== Convolution list dump ===\n");
	std::for_each(convolution_list.begin(), convolution_list.end(),
		std::mem_fun_ref(&TConvolutionSize::Dump));
}

/*
 *	bool TConvolutionList::Create(int n_classifier_num, const TWeakClassifier *p_classifier)
 *		- creates list of convolutions necessary to evaluate given classifiers
 *		- n_classifier_num is number of classifiers, p_classifier is list of classifiers
 *		- returns true on success, false on failure
 */
bool TConvolutionList::Create(int n_classifier_num, const TWeakClassifier *p_classifier)
{
	convolution_list.erase(convolution_list.begin(), convolution_list.end());
	// erase original contents if any

	for(int i = 0; i < n_classifier_num; ++ i) {
		TConvolutionSize t_conv_info(p_classifier[i]);
		if(std::find(convolution_list.begin(), convolution_list.end(),
		   t_conv_info) == convolution_list.end()) {
			convolution_list.reserve(convolution_list.size() + 1);
			if(convolution_list.capacity() < convolution_list.size() + 1)
				return false;
			convolution_list.push_back(t_conv_info);
		}
	}
	// simply create list of all different convolutions and shifts

	// todo - gather different convolutions, generate all possible shifts

	std::sort(convolution_list.begin(), convolution_list.end());
	// sort them by size

	return true;
}

/*
 *								=== ~TConvolutionList ===
 */

/*
 *								=== TShaderParameters ===
 */

TShaderParameters::TShaderParameters(CGLState *p_state, const TConvolutionSize &r_t_info,
	int n_x_shift, int n_y_shift, int n_src_w, int n_src_h)
	:n_source_w(n_src_w), n_source_h(n_src_h), n_offset_x(n_x_shift),
	n_offset_y(n_y_shift), n_samples_x(r_t_info.n_block_width),
	n_samples_y(r_t_info.n_block_height), n_mip_level(0)
{
	n_slice = n_x_shift + n_y_shift * r_t_info.n_block_width;
	// 3D tex slice

	while(!(n_offset_x & 1) && !(n_offset_y & 1) &&
	   !(n_samples_x & 1) && !(n_samples_y & 1) &&
	   !(n_source_w & 1) && !(n_source_h & 1)) {
		++ n_mip_level;
		n_offset_x /= 2;
		n_offset_y /= 2;
		n_samples_x /= 2;
		n_samples_y /= 2;
		n_source_w /= 2;
		n_source_h /= 2;
	}

	n_result_w = n_source_w / n_samples_x;
	n_result_w -= n_result_w % 4; // grey to RGBA pack
	n_result_h = n_source_h / n_samples_y;

	f_pixel_w = 1.0f / n_source_w;
	f_pixel_h = 1.0f / n_source_h;
	// calculate pixel size

	n_result_w /= 4;
	// grey to RGBA pack

	f_min_s = (n_offset_x + .25f) * f_pixel_w;
	f_min_t = (n_offset_y + .25f) * f_pixel_h; // sub-kernel offset (+ .25f to pixel center)
	f_max_s = f_min_s + (n_result_w - 1) * n_samples_x * 4 * f_pixel_w;
	f_max_t = f_min_t + (n_result_h - 1) * n_samples_y * f_pixel_h; // texture size is 1, 1
	// calculate texture coordinates

	int n_source_size = (n_source_w > n_source_h)? n_source_w : n_source_h;
	CRasterizerCalibrator::TTexCoordRect t_texcoords(f_min_s, f_min_t, f_max_s, f_max_t);
	if(CRasterizerCalibrator::CalibrateRasterizer(p_state, n_result_w, n_result_h,
	   t_texcoords, t_texcoords, .25f / n_source_size)) {
		f_min_s = t_texcoords.f_min_s;
		f_min_t = t_texcoords.f_min_t;
		f_max_s = t_texcoords.f_max_s;
		f_max_t = t_texcoords.f_max_t;

		n_source_w = n_src_w;
		n_source_h = n_src_h;
		// remember original source dimensions
	} else {
		n_source_w = -1;
		n_source_h = -1;
		// set invalid source dimensions to mark error
	}
	// calibrate rasterizer

	n_shader_type = (n_samples_x == 1 && n_samples_y == 1)? conv_PackShader : conv_ConvShader;
	// determine shader type
}

void TShaderParameters::Dump()
{
	printf("\tshader_params(type: \'%s\', offset-x: %d offset-y: %d,"
		" mip-level: %d, samples-x:%d, samples-y: %d, slice: %d)\n",
		(n_shader_type == conv_PackShader)? "pack" : "conv", n_offset_x, n_offset_y,
		n_mip_level, n_samples_x, n_samples_y, n_slice);
}

/*
 *								=== ~TShaderParameters ===
 */

TShaderParameters2::TShaderParameters2(const TShaderParameters &r_t_params,
	CGLShaderObject *_p_shader)
	:n_mip_level(r_t_params.n_mip_level), n_slice(r_t_params.n_slice),
	n_source_w(r_t_params.n_source_w), n_source_h(r_t_params.n_source_h), p_shader(_p_shader)
{
	bool b_result = true;
	if(r_t_params.n_shader_type == TShaderParameters::conv_PackShader) {
		if(!_p_shader->t_Find_Parameter("v_block_shift").SetValue(r_t_params.f_pixel_w,
		   2 * r_t_params.f_pixel_w, 3 * r_t_params.f_pixel_w, 0))
			b_result = false;
	} else {
		if(!_p_shader->t_Find_Parameter("v_block_shift").SetValue(r_t_params.f_pixel_w *
		   r_t_params.n_samples_x, 2 * r_t_params.n_samples_x * r_t_params.f_pixel_w,
		   3 * r_t_params.n_samples_x * r_t_params.f_pixel_w, 0) ||
		   !_p_shader->t_Find_Parameter("v_pixel").SetValue(r_t_params.f_pixel_w,
		   r_t_params.f_pixel_h) || !_p_shader->t_Find_Parameter("f_rescale").SetValue(1.0f /
		   (r_t_params.n_samples_x * r_t_params.n_samples_y)) ||
		   !_p_shader->t_Find_IntParameter("v_kernel").SetValue(r_t_params.n_samples_x,
		   r_t_params.n_samples_y))
			b_result = false;
	}
	p_params = (b_result)? _p_shader->p_GetParameterBuffer() : 0;

	p_quad_verts[0] = r_t_params.f_min_s;
	p_quad_verts[1] = r_t_params.f_max_t;
	p_quad_verts[2] = -1;
	p_quad_verts[3] =  1;
	p_quad_verts[4] =  0;
	p_quad_verts[5] = r_t_params.f_max_s;
	p_quad_verts[6] = r_t_params.f_max_t;
	p_quad_verts[7] =  1;
	p_quad_verts[8] =  1;
	p_quad_verts[9] =  0;
	p_quad_verts[10] = r_t_params.f_max_s;
	p_quad_verts[11] = r_t_params.f_min_t;
	p_quad_verts[12] =  1;
	p_quad_verts[13] = -1;
	p_quad_verts[14] =  0;
	p_quad_verts[15] = r_t_params.f_min_s;
	p_quad_verts[16] = r_t_params.f_min_t;
	p_quad_verts[17] = -1;
	p_quad_verts[18] = -1;
	p_quad_verts[19] =  0;
	// draw quad using interleaved arrays (GL_T2F_V3F)

	/*printf("\tparams2(this: 0x%08x, shader: 0x%08x)\n", this, p_shader);
	for(int i = 0; i < 4; ++ i) {
		printf("\t\tvertex(%g, %g, %g, %g, %g)\n", p_quad_verts[i * 5 + 0],
			p_quad_verts[i * 5 + 1], p_quad_verts[i * 5 + 2],
			p_quad_verts[i * 5 + 3], p_quad_verts[i * 5 + 4]);
	}*/ // debug dump
}

/*
 *								=== CConvolutionManager::CFindGLResource ===
 */

class CConvolutionManager::CFindGLResource {
protected:
	GLenum m_n_internal_format;
	int m_n_width, m_n_height;

public:
	CFindGLResource(int n_width, int n_height, GLenum n_internal_format)
		:m_n_width(n_width), m_n_height(n_height), m_n_internal_format(n_internal_format)
	{}

	bool operator ()(const CGLTexture_2D *p_texture) const
	{
		return p_texture->n_Pixel_Format() == m_n_internal_format &&
			p_texture->n_Width() == m_n_width &&
			p_texture->n_Height() == m_n_height;
	}

	bool operator ()(const CGLRenderBuffer *p_fbo) const
	{
		return p_fbo->n_Color_Format() == m_n_internal_format &&
			p_fbo->n_Width() == m_n_width &&
			p_fbo->n_Height() == m_n_height;
	}
};

/*
 *								=== ~CConvolutionManager::CFindGLResource ===
 */

/*
 *								=== CConvolutionManager ===
 */

CConvolutionManager::CConvolutionManager()
	:m_p_pack_shader(0), m_p_convolve_shader(0), m_b_status(false)
{}

CConvolutionManager::~CConvolutionManager()
{
	if(m_p_pack_shader)
		delete m_p_pack_shader;
	if(m_p_convolve_shader)
		delete m_p_convolve_shader;
	std::for_each(m_render_buffer_list.begin(), m_render_buffer_list.end(), Delete_FBO);
	std::for_each(m_image_list.begin(), m_image_list.end(), Delete_Texture);
}

bool CConvolutionManager::CreateConvolutionPlan(CGLState *p_state,
	const TConvolutionList &r_t_conv_list, int n_source_width, int n_source_height,
	bool b_16_bit_sums)
{
	m_shader_params_list.erase(m_shader_params_list.begin(), m_shader_params_list.end());
	// delete shader parameters

	std::for_each(m_render_buffer_list.begin(), m_render_buffer_list.end(), Delete_FBO);
	m_render_buffer_list.erase(m_render_buffer_list.begin(), m_render_buffer_list.end());
	std::for_each(m_image_list.begin(), m_image_list.end(), Delete_Texture);
	m_image_list.erase(m_image_list.begin(), m_image_list.end());
	// delete FBO's and textures

	m_b_status = false;

	if(!m_p_pack_shader) {
		TShaderInfo pack_shader("pack_shader", true,
			"uniform sampler2D n_texture;\n"
			"uniform vec4 v_block_shift;\n" // pixel-w, 2 * pixel-w, 3 * pixel-w, 0
			"\n"
			"void main()\n"
			"{\n"
			"    vec4 v_sample;\n"
			"    v_sample.x = texture2D(n_texture, gl_TexCoord[0].xy).x;\n"
			"    v_sample.y = texture2D(n_texture, gl_TexCoord[0].xy + v_block_shift.xw).x;\n"
			"    v_sample.z = texture2D(n_texture, gl_TexCoord[0].xy + v_block_shift.yw).x;\n"
			"    v_sample.w = texture2D(n_texture, gl_TexCoord[0].xy + v_block_shift.zw).x;\n" // pack four grey pixels to RGBA pixel
			"    gl_FragColor = v_sample;\n"
			"}\n");
		if(!(m_p_pack_shader = new CGLShaderObject(0, &pack_shader)))
			return false;
		char *p_s_compile = 0, *p_s_link = 0;
		bool b_result = m_p_pack_shader->Compile_Link(p_state, p_s_compile, p_s_link);
		if(p_s_compile) {
			printf("\n=== Pack shader compiler report ===\n%s\n", p_s_compile);
			delete[] p_s_compile;
		}
		if(p_s_link) {
			printf("\n=== Pack shader linker report ===\n%s\n", p_s_link);
			delete[] p_s_link;
		}
		if(!b_result) {
			delete m_p_pack_shader;
			m_p_pack_shader = 0;
			return false;
		}
	}
	if(!m_p_convolve_shader) {
		TShaderInfo convolve_shader("convolve_shader", true,
			"uniform sampler2D n_texture;\n"
			"uniform vec2 v_pixel;\n" // pixel-w, pixel-h
			"uniform ivec2 v_kernel;\n" // width, height
			"uniform vec4 v_block_shift;\n" // width * pixel-w, 2 * width * pixel-w, 3 * width * pixel-w, 0
			"uniform float f_rescale;\n" // 1.0f / (width * height)
			"\n"
			"void main()\n"
			"{\n"
			"    vec4 v_sum = vec4(.0, .0, .0, .0);\n"
			"    for(int i = 0; i != v_kernel.x; ++ i) {\n"
			"        for(int j = 0; j != v_kernel.y; ++ j) {\n"
			"            vec2 v_tc = gl_TexCoord[0].xy + v_pixel * vec2(i, j);\n"
			"            vec4 v_sample;\n"
			"            v_sample.x = texture2D(n_texture, v_tc).x;\n"
			"            v_sample.y = texture2D(n_texture, v_tc + v_block_shift.xw).x;\n"
			"            v_sample.z = texture2D(n_texture, v_tc + v_block_shift.yw).x;\n"
			"            v_sample.w = texture2D(n_texture, v_tc + v_block_shift.zw).x;\n" // calculate sum of four blocks at once
			"            v_sum += v_sample\n;"
			"        }\n"
			"    }\n"
			"    gl_FragColor = v_sum * f_rescale;\n"
			"}\n");
		if(!(m_p_convolve_shader = new CGLShaderObject(0, &convolve_shader)))
			return false;
		char *p_s_compile = 0, *p_s_link = 0;
		bool b_result = m_p_convolve_shader->Compile_Link(p_state, p_s_compile, p_s_link);
		if(p_s_compile) {
			printf("\n=== Convolve shader compiler report ===\n%s\n", p_s_compile);
			delete[] p_s_compile;
		}
		if(p_s_link) {
			printf("\n=== Convolve shader linker report ===\n%s\n", p_s_link);
			delete[] p_s_link;
		}
		if(!b_result) {
			delete m_p_convolve_shader;
			m_p_convolve_shader = 0;
			return false;
		}
	}
	// create shaders

	m_image_list.reserve(r_t_conv_list.convolution_list.size());
	if(m_image_list.capacity() < r_t_conv_list.convolution_list.size())
		return false;
	// reserve enough space for textures

	std::string s_vendor((const char*)glGetString(GL_VENDOR));
	for(int i = s_vendor.length() - 1; i >= 0; -- i) {
		if(!isalpha(s_vendor[i]))
			s_vendor.erase(s_vendor.begin() + i, s_vendor.begin() + i + 1);
		else
			s_vendor[i] = tolower(s_vendor[i]);
	}
	m_b_render_to_3d_tex = s_vendor.find("nvidia") == std::string::npos ||
		strstr((const char*)glGetString(GL_RENDERER), "8800");
	if(m_b_render_to_3d_tex)
		printf("\tnote: using direct render to 3D texture\n");
	else {
		printf("\twarning: render to 3D texture via texture copy"
			" (expect less than a half of potential speed)\n");
	}
	// nVidia bug with direct render to 3D texture using
	// render buffers (reportedly working on 8800s)

	for(int i = 0; i < r_t_conv_list.convolution_list.size(); ++ i) {
		TConvolutionSize t_convolution = r_t_conv_list.convolution_list[i];
		// pick convolution

		int n_result_width = n_source_width / t_convolution.n_block_width / 4;
		int n_result_height = n_source_height / t_convolution.n_block_height;
		int n_result_depth = t_convolution.n_block_width * t_convolution.n_block_height;
		// determine output size

		CGLRenderBuffer_FBO *p_render_buffer;
		std::vector<CGLRenderBuffer_FBO*>::const_iterator p_rb_iter;
		if((p_rb_iter = std::find_if(m_render_buffer_list.begin(), m_render_buffer_list.end(),
		   CFindGLResource(n_result_width, n_result_height, (b_16_bit_sums)?
		   GL_RGBA16 : GL_RGBA8))) == m_render_buffer_list.end()) {
			m_render_buffer_list.reserve(m_render_buffer_list.size() + 1);
			if(m_render_buffer_list.capacity() < m_render_buffer_list.size() + 1)
				return false;
			// make sure there is enough space in the list

			CGLRenderBuffer_FBO *_p_render_buffer;
			if(!(_p_render_buffer = new CGLRenderBuffer_FBO(n_result_width,
			   n_result_height, 1, m_b_render_to_3d_tex, (b_16_bit_sums)? GL_RGBA16 : GL_RGBA8,
			   false, false, 0, false, false, 0)))
				return false;
			if(!_p_render_buffer->b_Status() || glGetError() != GL_NO_ERROR) {
				delete _p_render_buffer;
				return false;
			}
			// create a new render buffer

			p_render_buffer = _p_render_buffer;
			m_render_buffer_list.push_back(_p_render_buffer);
		} else
			p_render_buffer = *p_rb_iter;
		// try to find suitable framebuffer or create one

		CGLTexture_3D *p_texture;
		if(!(p_texture = new CGLTexture_3D(p_state, n_result_width, n_result_height,
		   n_result_depth, (b_16_bit_sums)? GL_RGBA16 : GL_RGBA8, false, 0)))
			return false;
		p_texture->r_Parameters(p_state).Set_Texture_Wrap_S(GL_REPEAT/*GL_CLAMP_TO_BORDER*/);
		p_texture->r_Parameters(p_state).Set_Texture_Wrap_T(GL_REPEAT/*GL_CLAMP_TO_BORDER*/);
		p_texture->r_Parameters(p_state).Set_Texture_Wrap_R(GL_REPEAT); // no R clamping
		p_texture->r_Parameters(p_state).Set_Texture_Mag_Filter(GL_NEAREST);
		p_texture->r_Parameters(p_state).Set_Texture_Min_Filter(GL_NEAREST); // need no mipmaps
		if(!p_texture->b_Status() || glGetError() != GL_NO_ERROR) {
			delete p_texture;
			return false;
		}
		m_image_list.push_back(TImageData(t_convolution, p_texture)); // enough space reserved before entering this cycle
		// create texture to hold convolution results

		m_shader_params_list.reserve(m_shader_params_list.size() + n_result_depth);
		if(m_shader_params_list.capacity() < m_shader_params_list.size() + n_result_depth)
			return false;
		// there will be different shader parameters to calculate each 3D texture slice

		for(int n_y_shift = 0; n_y_shift < t_convolution.n_block_height; ++ n_y_shift) {
			for(int n_x_shift = 0; n_x_shift < t_convolution.n_block_width; ++ n_x_shift) {
				TShaderParameters t_sh_p(p_state, t_convolution, n_x_shift, n_y_shift,
					n_source_width, n_source_height);
				if(t_sh_p.n_source_w < 0)
					return false; // failed to calibrate
				// create shader parameters

				t_sh_p.Dump();

				TShaderParameters2 t_sh_p2(t_sh_p, (t_sh_p.n_shader_type ==
					TShaderParameters::conv_PackShader)? m_p_pack_shader : m_p_convolve_shader);
				if(!t_sh_p2.p_params)
					return false; // not enough memory for parameters
				t_sh_p2.p_render_buffer = p_render_buffer;
				t_sh_p2.p_image = p_texture;
				// create buffered shader parameters

				_ASSERTE(t_sh_p2.n_slice == n_x_shift + n_y_shift * t_convolution.n_block_width);
				_ASSERTE(m_image_list.back().n_Slice(n_x_shift, n_y_shift) == t_sh_p2.n_slice);

				m_shader_params_list.push_back(t_sh_p2);
			}
		}
		// create shader parameters for this convolution
	}
	// create list with shader parameters

	m_b_status = true;
	m_n_source_w = n_source_width;
	m_n_source_h = n_source_height;
	m_convolution_list = r_t_conv_list.convolution_list; // validation purposes

	return true;
}

bool CConvolutionManager::Convolve(CGLState *p_state, CGLTexture_2D *p_source_image)
{
	if(!m_b_status)
		return false;
	// shaders aren't ready

	if(m_n_source_w != p_source_image->n_Width() ||
	   m_n_source_h != p_source_image->n_Height())
		return false;
	// convolution plan for different image size

	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// reset matrices

	p_state->DisableTexture3D();
	p_source_image->r_Parameters(p_state).Set_Texture_Min_Filter(GL_NEAREST);
	p_source_image->r_Parameters(p_state).Set_Texture_Mag_Filter(GL_NEAREST);
	p_source_image->r_Parameters(p_state).Set_Texture_Wrap_S(GL_CLAMP_TO_BORDER);
	p_source_image->r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP_TO_BORDER);
	// make sure the texture has correct filtering

	int p_orig_viewport[4];
	glGetIntegerv(GL_VIEWPORT, p_orig_viewport);

	bool b_framebuffer_errors = false;
	CGLRenderBuffer *p_prev_rb = 0;
	for(int i = 0; i < m_shader_params_list.size(); ++ i) {
		TShaderParameters2 t_params = m_shader_params_list[i];

		_ASSERTE(std::find(m_render_buffer_list.begin(), m_render_buffer_list.end(),
			t_params.p_render_buffer) != m_render_buffer_list.end());
		_ASSERTE(std::find(m_image_list.begin(), m_image_list.end(),
			t_params.p_image) != m_image_list.end());

		if(p_prev_rb && p_prev_rb != t_params.p_render_buffer)
			p_prev_rb->Release();
		p_prev_rb = t_params.p_render_buffer;
		t_params.p_render_buffer->Bind();
		// bind render buffer in case we're using another one

		if(m_b_render_to_3d_tex)
			t_params.p_render_buffer->BindColorTexture_3D(*t_params.p_image, t_params.n_slice, 0);
		glViewport(0, 0, t_params.p_render_buffer->n_Width(),
			t_params.p_render_buffer->n_Height());
		// bind texture for rendering

		if(!t_params.p_render_buffer->b_Status()) {
			b_framebuffer_errors = true;
			break;
		}
		// check render buffer status

		//p_state = t_params.p_render_buffer->p_State(); // we're using FBO's, they don't have it's own state
		p_state->DisableTexture3D();
		p_source_image->Bind_Enable(p_state);
		p_source_image->r_Parameters(p_state).Set_Texture_Base_Level(t_params.n_mip_level);
		p_source_image->r_Parameters(p_state).Set_Texture_Max_Level(t_params.n_mip_level);
		p_state->DisableDepthTest();
		p_state->DisableCullFace();
		p_state->DisableBlend();
		p_state->DisableAlphaTest();
		//p_state->Color3f(1, 1, 1); // doesn't matter, shaders ignore it
		// setup GL state

		//t_params.p_shader->SetParameterBuffer(*t_params.p_params);
		t_params.p_shader->Bind(p_state, t_params.p_params); // without copying
		// bind shader

		glInterleavedArrays(GL_T2F_V3F, 0, t_params.p_quad_verts);
		glDrawArrays(GL_QUADS, 0, 4);
			// draw fullscreen quad to execute shader

		if(m_b_render_to_3d_tex)
			t_params.p_render_buffer->BindColorTexture_3D(0, 0, 0);
		else {
			int n_read_buffer;
			int n_fbo_buffer = t_params.p_render_buffer->n_Draw_Buffer(0);
			glGetIntegerv(GL_READ_BUFFER, &n_read_buffer);
			if(n_fbo_buffer != n_read_buffer)
				glReadBuffer(n_fbo_buffer);
			// make sure we're reading from FBO

			p_state->DisableTexture2D();
			t_params.p_image->Bind_Enable(p_state);
			glCopyTexSubImage3D(GL_TEXTURE_3D, 0, 0, 0, t_params.n_slice,
				0, 0, t_params.p_image->n_Width(), t_params.p_image->n_Height());
			// manually copy texture slice

			if(n_fbo_buffer != n_read_buffer)
				glReadBuffer(n_read_buffer);
			// put back the original read-buffer
		}
		// release texture
	}

	p_prev_rb->Release();

	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);
	// enabled by InterleavedArrays

	p_state->DisableTexture3D();
	p_source_image->Bind_Enable(p_state);
	p_source_image->r_Parameters(p_state).Set_Texture_Base_Level(0);
	p_source_image->r_Parameters(p_state).Set_Texture_Max_Level(1000);
	// put back normal mip level

	if(m_p_pack_shader->n_Api() == shader_api_Core_shader)
		p_state->BindProgramObject_GL2(0);
	else
		p_state->BindProgramObject(0);
	// bind no shader

	glViewport(p_orig_viewport[0], p_orig_viewport[1], p_orig_viewport[2], p_orig_viewport[3]);
	// restore original viewport

	return !b_framebuffer_errors && glGetError() == GL_NO_ERROR;
}

#pragma pack(1)

struct TTgaHeader {
	unsigned char n_id_length;
	unsigned char n_palette_type;
	unsigned char n_image_type;
	unsigned short n_first_color;
	unsigned short n_palette_colors;
	unsigned char n_palette_entry_size;
	unsigned short n_left;
	unsigned short n_top;
	unsigned short n_image_width;
	unsigned short n_image_height;
	unsigned char n_bpp;
	unsigned char n_descriptor_bits;
};

#pragma pack()

/*
 *	bool Save_Gray_TGA(const char *p_s_file, int n_width, int n_height,
 *		const unsigned char *p_buffer)
 *		- write grayscale targa image to file p_s_file
 *		- returns true on success, false on failure
 */
bool Save_Gray_TGA(const char *p_s_file, int n_width, int n_height, const unsigned char *p_buffer)
{
	FILE *p_fw;
	if(!(p_fw = fopen(p_s_file, "wb")))
		return false;

	TTgaHeader t_header = {0};
	//t_header.n_id_length = 0;
	t_header.n_bpp = 8;
	t_header.n_image_type = 3 | 8; // grayscale, compressed
	//t_header.n_palette_type = 0;
	//t_header.n_palette_entry_size = 8;
	//t_header.n_first_color = 0;
	//t_header.n_palette_colors = 0;
	//t_header.n_left = 0;
	//t_header.n_top = 0;
	t_header.n_image_width = n_width;
	t_header.n_image_height = n_height;
	t_header.n_descriptor_bits = 0x20; // upside-down
	// create header

	if(fwrite(&t_header, sizeof(t_header), 1, p_fw) != 1) {
		fclose(p_fw);
		return false;
	}
	// write header

	for(const unsigned char *p_scanline = p_buffer, *p_end = p_buffer +
	   (n_width * n_height); p_scanline != p_end; p_scanline += n_width) {
		for(const unsigned char *p_ptr = p_scanline, *p_end2 = p_scanline + n_width;
		   p_ptr != p_end2; ++ p_ptr) {
			int n_remains = p_end2 - p_ptr;
			if(n_remains > 128)
				n_remains = 128;
			bool b_compress = n_remains > 2 && /*n_Gray*/(*p_ptr) == /*n_Gray*/(*(p_ptr + 1)) &&
				/*n_Gray*/(*p_ptr) == /*n_Gray*/(*(p_ptr + 2));
			int n_run_length = n_remains;
			for(int i = 1; i < n_remains; ++ i) {
				if(b_compress && /*n_Gray*/(*p_ptr) != /*n_Gray*/(*(p_ptr + i))) {
					n_run_length = i;
					break;
				} else if(!b_compress && i + 1 < n_remains &&
				   /*n_Gray*/(*(p_ptr + i - 1)) == /*n_Gray*/(*(p_ptr + i)) &&
				   /*n_Gray*/(*(p_ptr + i - 1)) == /*n_Gray*/(*(p_ptr + i + 1))) {
					n_run_length = i;
					break;
				}
			}
			// determine run length and wheter to compress data

			unsigned char n_code = (n_run_length - 1) | ((b_compress)? 0x80 : 0x00);
			if(fwrite(&n_code, sizeof(char), 1, p_fw) != 1) {
				fclose(p_fw);
				return false;
			}

			if(b_compress) {
				unsigned char n_gray = *p_ptr;
				p_ptr += n_run_length - 1;
				if(fwrite(&n_gray, sizeof(char), 1, p_fw) != 1) {
					fclose(p_fw);
					return false;
				}
			} else {
				for(const unsigned char *p_end2 = p_ptr + n_run_length;
				   p_ptr != p_end2; ++ p_ptr) {
					if(fwrite(p_ptr, sizeof(char), 1, p_fw) != 1) {
						fclose(p_fw);
						return false;
					}
				}
				-- p_ptr; // ++ in the for command
			}
		}
	}
	// write data

	fclose(p_fw);

	return true;
}

bool CConvolutionManager::Validate(CGLState *p_state, CGLTexture_2D *p_source_image,
	int &r_n_max_difference, float &r_f_standard_deviation, bool b_save_images)
{
	r_n_max_difference = 0;
	r_f_standard_deviation = 0;

	if(!Convolve(p_state, p_source_image))
		return false;

	unsigned char *p_image_buffers;
	int n_image_width = p_source_image->n_Width();
	int n_image_height = p_source_image->n_Height();
	int n_image_size = n_image_width * n_image_height;
	if(!(p_image_buffers = new unsigned char[n_image_size * 3]))
		return false;
	// need image buffer for original image (so CPU convolver has a source data),
	// buffer for CPU convolution and for GPU convolution

	unsigned char *p_src_buffer = p_image_buffers;
	unsigned char *p_cpu_buffer = p_image_buffers + n_image_size;
	unsigned char *p_gpu_buffer = p_image_buffers + 2 * n_image_size;
	// pointers to the image buffers (convolutions are up to size of original image)

	glPixelStorei(GL_PACK_ALIGNMENT, 1);
	// byte-aligned pixels

	p_state->DisableTexture3D();
	p_source_image->Bind_Enable(p_state);
	glGetTexImage(GL_TEXTURE_2D, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, p_src_buffer);
	// get source image data

	int n_deviation_num = 0;

	std::vector<TShaderParameters2>::const_iterator p_shader_params = m_shader_params_list.begin();
	for(int i = 0; i < m_convolution_list.size(); ++ i) {
		int n_block_width = m_convolution_list[i].n_block_width;
		int n_block_height = m_convolution_list[i].n_block_height;
		int n_block_size = n_block_width * n_block_height;

		p_gpu_buffer = p_image_buffers + 2 * n_image_size;
		// pointer is shifted in the cycle below

		p_state->DisableTexture2D();
		(*p_shader_params).p_image->Bind_Enable(p_state);
		glGetTexImage(GL_TEXTURE_3D, 0, GL_RGBA, GL_UNSIGNED_BYTE, p_gpu_buffer);
		// get all the slices of this image

		int n_slice_width = (*p_shader_params).p_image->n_Width() * 4;
		int n_slice_height = (*p_shader_params).p_image->n_Height();
		int n_slice_size = n_slice_width * n_slice_height;
		// slices are RGBA packed so their real width is 4 times larger

		for(int n_y_shift = 0; n_y_shift < n_block_height; ++ n_y_shift) {
			for(int n_x_shift = 0; n_x_shift < n_block_width; ++ n_x_shift,
			   ++ p_shader_params, p_gpu_buffer += n_slice_size) {
				_ASSERTE(p_shader_params != m_shader_params_list.end());

				for(int y = 0; y < n_slice_height; ++ y) {
					for(int x = 0; x < n_slice_width; ++ x) {
						int n_win_x0 = x * n_block_width + n_x_shift;
						int n_win_y0 = y * n_block_height + n_y_shift;
						int n_win_x1 = n_win_x0 + n_block_width;
						int n_win_y1 = n_win_y0 + n_block_height;
						if(n_win_x1 > n_image_width)
							n_win_x1 = n_image_width;
						if(n_win_y1 > n_image_height)
							n_win_y1 = n_image_height;
						int n_sum = 0;
						for(int n_y = n_win_y0; n_y < n_win_y1; ++ n_y) {
							for(int n_x = n_win_x0; n_x < n_win_x1; ++ n_x)
								n_sum += p_src_buffer[n_x + n_y * n_image_width];
						}
						n_sum += n_block_size / 2; // round so we get closer to GPU
						n_sum /= n_block_size;
						p_cpu_buffer[x + y * n_slice_width] = n_sum;
					}
				}
				// calculate convolution on CPU

				int n_max_difference = 0;
				float f_sum_difference2 = 0;
				for(int y = 0; y < n_slice_height; ++ y) {
					for(int x = 0; x < n_slice_width; ++ x) {
						int n_diff = abs(p_cpu_buffer[x + y * n_slice_width] -
							p_gpu_buffer[x + y * n_slice_width]);
						if(n_max_difference < n_diff)
							n_max_difference = n_diff;
						f_sum_difference2 += n_diff * n_diff;
					}
				}
				f_sum_difference2 *= 1.0f / n_slice_size;
				// calculate maximal difference and standard deviation

				if(r_n_max_difference < n_max_difference)
					r_n_max_difference = n_max_difference;
				r_f_standard_deviation += f_sum_difference2;
				n_deviation_num ++;//= n_slice_size;
				// update statistic variables

				if(n_max_difference > 5) {
					fprintf(stderr, "\terror: convolution didn't pass the validation"
						" on slice %d; max error %d, standard deviation %g\n",
						p_shader_params - m_shader_params_list.begin(),
						n_max_difference, sqrt(f_sum_difference2));
					// prints error

					if(b_save_images) {
						char p_s_filename[64];
						sprintf(p_s_filename, "im_slice_%02d_gpu.tga",
							p_shader_params - m_shader_params_list.begin());
						if(!Save_Gray_TGA(p_s_filename,
						   n_slice_width, n_slice_height, p_gpu_buffer)) {
							fprintf(stderr, "\terror: error writing debug"
								" image \'%s\'\n", p_s_filename);
						}
						sprintf(p_s_filename, "im_slice_%02d_cpu.tga",
							p_shader_params - m_shader_params_list.begin());
						if(!Save_Gray_TGA(p_s_filename,
						   n_slice_width, n_slice_height, p_cpu_buffer)) {
							fprintf(stderr, "\terror: error writing reference"
								" image \'%s\'\n", p_s_filename);
						}
					}
					// save images for easier debugging

					r_f_standard_deviation = (float)sqrt(r_f_standard_deviation / n_deviation_num);
					// calculate standard deviation

					delete[] p_image_buffers;

					return false;
				}
			}
		}
	}
	// go trough all convolution images and look for differences with the ones obtained using CPU

	delete[] p_image_buffers;

	r_f_standard_deviation = (float)sqrt(r_f_standard_deviation / n_deviation_num);
	// calculate standard deviation

	return true;
}

void CConvolutionManager::Delete_FBO(CGLRenderBuffer_FBO *p_fbo)
{
	delete p_fbo;
}

void CConvolutionManager::Delete_Texture(TImageData &r_t_data)
{
	delete r_t_data.p_texture;
}

/*
 *								=== ~CConvolutionManager ===
 */
