/*
								+-----------------------------------+
								|                                   |
								|  ***  Convolution scheduler  ***  |
								|                                   |
								|   Copyright   -tHE SWINe- 2007   |
								|                                   |
								|            ConvPlan.h             |
								|                                   |
								+-----------------------------------+
*/

#ifndef __ADABOOST_CONVOLUTION_SCHEDULER_INCLUDED
#define __ADABOOST_CONVOLUTION_SCHEDULER_INCLUDED

/*
 *	in order to achieve good memory throughput, packed pixels optimalization is introduced.
 *	quadruples of greyscale pixels are stored as RGBA values so reading any four horizontaly
 *	neighbouring values requires one to two texture samples.
 *
 *	LRD's are calculated over grids of 3x3 (or 4x4) pixels and aim is to read those
 *	pixels using fewest texture samples possible. each pixel of such a grid contains result
 *	of convolution (more precisely sum of pixels under rectangular window) of certain size.
 *
 *	in order to have grid pixels next to each other, it's necessary to calculate smaller
 *	version of input image containing interleaved convolutions. but then, on the other side
 *	pixel coordinates are rounded to convolution sizes so it's necessary to have more versions
 *	of those images, each containing interleaved convolutions of original image with different
 *	shift.
 *
 *	the purpose of convolution scheduler is to determine convolution sizes and offsets required
 *	to evaluate set of given weak classifiers.
 *
 *	offset is always directed inside the image, ie. for example for block size 2x2 and image
 *	size 256x256, these convolutions can be created:
 *
 *	block 2x2, shift (0, 0): 128x128 image created from rows 0-255 and cols 0-255
 *
 *	block 2x2, shift (1, 0): 128x128 image created from rows 0-255 and cols 1-255 and an extra
 *	empty column
 *
 *	block 2x2, shift (0, 1): 128x128 image created from rows 1-255 and an extra empty row and
 *	cols 0-255
 *
 *	block 2x2, shift (1, 1): 128x128 image created from rows 1-255 and an extra empty row and
 *	cols 1-255 and an extra empty column
 *
 *	TWeakClassifier::n_position_x and TWeakClassifier::n_position_y are then divided by block
 *	size to yield correct offset in convolution image, whereas their modulo division yield
 *	necessary image shift.
 */

struct TConvolutionSize {
	int n_block_width;
	int n_block_height; // block size in pixels

	/*
	 *	TConvolutionSize::TConvolutionSize()
	 *		- default constructor
	 */
	TConvolutionSize();

	/*
	 *	TConvolutionSize::TConvolutionSize(const TConvolutionSize &r_t_info)
	 *		- copy constructor
	 */
	TConvolutionSize(const TConvolutionSize &r_t_info);

	/*
	 *	TConvolutionSize::TConvolutionSize(const TWeakClassifier &r_t_weak_cl)
	 *		- constructor creating convolution info satisfying weak classifier demands
	 */
	TConvolutionSize(const TWeakClassifier &r_t_weak_cl);

	/*
	 *	TConvolutionSize TConvolutionSize::operator =(const TConvolutionSize &r_t_info)
	 *		- copy operator
	 */
	TConvolutionSize operator =(const TConvolutionSize &r_t_info);

	/*
	 *	bool TConvolutionSize::operator ==(const TConvolutionSize &r_t_info)
	 *		- equality operator
	 */
	bool operator ==(const TConvolutionSize &r_t_info);

	/*
	 *	bool TConvolutionSize::operator <(const TConvolutionSize &r_t_info)
	 *		- less-then operator for sorting
	 */
	bool operator <(const TConvolutionSize &r_t_info);

	/*
	 *	int TConvolutionSize::Dump()
	 *		- prints convolution info dump to stdout
	 *		- MSVC can't handle return void; in templates so it returns number of prited chars
	 */
	int Dump();
};

struct TConvolutionList {
	std::vector<TConvolutionSize> convolution_list;

	/*
	 *	void TConvolutionList::Dump()
	 *		- prints convolution info dumps to stdout
	 */
	void Dump();

	/*
	 *	bool TConvolutionList::Create(int n_classifier_num, const TWeakClassifier *p_classifier)
	 *		- creates list of convolutions necessary to evaluate given classifiers
	 *		- n_classifier_num is number of classifiers, p_classifier is list of classifiers
	 *		- returns true on success, false on failure
	 */
	bool Create(int n_classifier_num, const TWeakClassifier *p_classifier);
};

struct TShaderParameters {
	enum {
		conv_PackShader,
		conv_ConvShader
	};

	int n_shader_type;
	int n_offset_x, n_offset_y; // x and y offset of texture
	int n_mip_level, n_slice; // mip-map texture level, 3d texture slice
	int n_samples_x, n_samples_y; // number of horiz and vert samples
	float f_pixel_w, f_pixel_h; // pixel size
	float f_min_s, f_min_t, f_max_s, f_max_t; // min and max texcoords
	int n_source_w, n_source_h;
	int n_result_w, n_result_h;

	TShaderParameters(CGLState *p_state, const TConvolutionSize &r_t_info,
		int n_x_shift, int n_y_shift, int n_src_w, int n_src_h);

	void Dump();
};

struct TShaderParameters2 {
	CGLRenderBuffer *p_render_buffer;
	CGLTexture_3D *p_image;
	CGLShaderObject *p_shader;
	CGLShaderParams *p_params;
	int n_mip_level, n_slice;
	int n_source_w, n_source_h;
	float p_quad_verts[20];

	TShaderParameters2(const TShaderParameters &r_t_params, CGLShaderObject *_p_shader);
};

class CConvolutionManager {
public:
	struct TImageData : public TConvolutionSize {
		CGLTexture_3D *p_texture;

		TImageData(const TConvolutionSize &r_t_size, CGLTexture_3D *p_tex)
			:TConvolutionSize(r_t_size), p_texture(p_tex)
		{}

		int n_Slice(int n_x_shift, int n_y_shift) const
		{
			_ASSERTE(n_x_shift >= 0 && n_x_shift < n_block_width);
			_ASSERTE(n_y_shift >= 0 && n_y_shift < n_block_height);
			return n_x_shift + n_y_shift * n_block_width;
		}

		/*
		 *	bool TConvolutionSize::operator ==(const CGLTexture_3D *p_tex)
		 *		- equality operator
		 */
		bool operator ==(const CGLTexture_3D *p_tex)
		{
			return p_texture == p_tex;
		}
	};

protected:
	std::vector<TConvolutionSize> m_convolution_list;
	std::vector<CGLRenderBuffer_FBO*> m_render_buffer_list;
	CGLShaderObject *m_p_pack_shader, *m_p_convolve_shader;
	std::vector<TShaderParameters2> m_shader_params_list;

	std::vector<TImageData> m_image_list;
	int m_n_source_w, m_n_source_h;
	bool m_b_status, m_b_render_to_3d_tex;

	class CFindGLResource;

public:
	CConvolutionManager();
	~CConvolutionManager();

	bool CreateConvolutionPlan(CGLState *p_state, const TConvolutionList &r_t_conv_list,
		int n_source_width, int n_source_height, bool b_16_bit_sums = true);

	bool Convolve(CGLState *p_state, CGLTexture_2D *p_source_image);

	bool Validate(CGLState *p_state, CGLTexture_2D *p_source_image,
		int &r_n_max_difference, float &r_f_standard_deviation, bool b_save_images = true);

	inline int n_Convolution_Num() const
	{
		return m_image_list.size();
	}

	inline const TImageData &r_Convolution_Image(int n_index) const
	{
		return m_image_list[n_index];
	}

	inline TImageData &r_Convolution_Image(int n_index)
	{
		return m_image_list[n_index];
	}

protected:
	static void Delete_FBO(CGLRenderBuffer_FBO *p_fbo);
	static void Delete_Texture(TImageData &r_t_data);
};

/*
const char *p_s_shader_copy_image =
	"uniform sampler2D n_texture;\n"
	"uniform vec4 v_block_shift;\n" // pixel-w, 2 * pixel-w, 3 * pixel-w, 0
	"\n"
	"void main()\n"
	"{\n"
	"    vec4 v_sample;\n"
	"    v_sample.x = texture2D(n_texture, gl_TexCoord[0].xy).x;\n"
	"    v_sample.y = texture2D(n_texture, gl_TexCoord[0].xy + v_block_shift.xw).x;\n"
	"    v_sample.z = texture2D(n_texture, gl_TexCoord[0].xy + v_block_shift.yw).x;\n"
	"    v_sample.w = texture2D(n_texture, gl_TexCoord[0].xy + v_block_shift.zw).x;\n" // pack four grey pixels to RGBA pixel
	"    gl_FragColor = v_sample;\n"
	"}\n";

const char *p_s_shader_convolve_image =
	"uniform sampler2D n_texture;\n"
	"uniform vec2 v_pixel;\n" // pixel-w, pixel-h
	"uniform ivec2 v_kernel;\n" // width, height
	"uniform vec4 v_block_shift;\n" // width * pixel-w, 2 * width * pixel-w, 3 * width * pixel-w, 0
	"\n"
	"void main()\n"
	"{\n"
	"    vec4 v_sum = vec4(.0, .0, .0, .0);\n"
	"    vec2 v_texcoord = gl_TexCoord[0].xy;\n"
	"    for(int i = 0; i != v_kernel.x; ++ i) {\n"
	"        v_texcoord.x = gl_TexCoord[0].x;\n"
	"        for(int j = 0; j != v_kernel.y; ++ j) {\n"
    "            vec4 v_sample;\n"
	"            v_sample.x = texture2D(n_texture, v_texcoord).x;\n"
	"            v_sample.y = texture2D(n_texture, v_texcoord + v_block_shift.xw).x;\n"
	"            v_sample.z = texture2D(n_texture, v_texcoord + v_block_shift.yw).x;\n"
	"            v_sample.w = texture2D(n_texture, v_texcoord + v_block_shift.zw).x;\n" // calculate sum of four blocks at once
	"            v_texcoord.x += v_pixel.x;\n"
	"            v_sum += v_sample\n;"
	"        }\n"
	"        v_texcoord.y += v_pixel.y;\n"
	"    }\n"
	"    gl_FragColor = v_sum;\n"
	"}\n";
*/

#endif //__ADABOOST_CONVOLUTION_SCHEDULER_INCLUDED
