/*
								+-----------------------------------+
								|                                   |
								|  ***  Convolution scheduler  ***  |
								|                                   |
								|   Copyright   -tHE SWINe- 2007   |
								|                                   |
								|          ConvSchedule.h           |
								|                                   |
								+-----------------------------------+
*/

#ifndef __ADABOOST_CONVOLUTION_SCHEDULER_INCLUDED
#define __ADABOOST_CONVOLUTION_SCHEDULER_INCLUDED

/*
 *	in order to achieve good memory throughput, packed pixels optimalization is introduced.
 *	quadruples of greyscale pixels are stored as RGBA values so reading any four horizontaly
 *	neighbouring values requires one to two texture samples.
 *
 *	LRD's are calculated over grids of 3x3 (or 4x4) pixels and aim is to read those
 *	pixels using fewest texture samples possible. each pixel of such a grid contains result
 *	of convolution (more precisely sum of pixels under rectangular window) of certain size.
 *
 *	in order to have grid pixels next to each other, it's necessary to calculate smaller
 *	version of input image containing interleaved convolutions. but then, on the other side
 *	pixel coordinates are rounded to convolution sizes so it's necessary to have more versions
 *	of those images, each containing interleaved convolutions of original image with different
 *	shift.
 *
 *	the purpose of convolution scheduler is to determine convolution sizes and offsets required
 *	to evaluate set of given weak classifiers.
 *
 *	offset is always directed inside the image, ie. for example for block size 2x2 and image
 *	size 256x256, these convolutions can be created:
 *
 *	block 2x2, shift (0, 0): 128x128 image created from rows 0-255 and cols 0-255
 *
 *	block 2x2, shift (1, 0): 128x128 image created from rows 0-255 and cols 1-255 and an extra
 *	empty column
 *
 *	block 2x2, shift (0, 1): 128x128 image created from rows 1-255 and an extra empty row and
 *	cols 0-255
 *
 *	block 2x2, shift (1, 1): 128x128 image created from rows 1-255 and an extra empty row and
 *	cols 1-255 and an extra empty column
 *
 *	TWeakClassifier::n_position_x and TWeakClassifier::n_position_y are then divided by block
 *	size to yield correct offset in convolution image, whereas their modulo division yield
 *	necessary image shift.
 */

struct TConvolutionInfo {
	int n_block_width;
	int n_block_height; // block size in pixels

	int n_x_shift;
	int n_y_shift; // offset in image

	/*
	 *	TConvolutionInfo::TConvolutionInfo()
	 *		- default constructor
	 */
	TConvolutionInfo();

	/*
	 *	TConvolutionInfo::TConvolutionInfo(const TConvolutionInfo &r_t_info)
	 *		- copy constructor
	 */
	TConvolutionInfo(const TConvolutionInfo &r_t_info);

	/*
	 *	TConvolutionInfo::TConvolutionInfo(const TWeakClassifier &r_t_weak_cl)
	 *		- constructor creating convolution info satisfying weak classifier demands
	 */
	TConvolutionInfo(const TWeakClassifier &r_t_weak_cl);

	/*
	 *	TConvolutionInfo TConvolutionInfo::operator =(const TConvolutionInfo &r_t_info)
	 *		- copy operator
	 */
	TConvolutionInfo operator =(const TConvolutionInfo &r_t_info);

	/*
	 *	bool TConvolutionInfo::operator ==(const TConvolutionInfo &r_t_info)
	 *		- equality operator
	 */
	bool operator ==(const TConvolutionInfo &r_t_info);

	/*
	 *	bool TConvolutionInfo::operator <(const TConvolutionInfo &r_t_info)
	 *		- less-then operator for sorting
	 */
	bool operator <(const TConvolutionInfo &r_t_info);

	/*
	 *	int TConvolutionInfo::Dump()
	 *		- prints convolution info dump to stdout
	 *		- MSVC can't handle return void; in templates so it returns number of prited chars
	 */
	int Dump();
};

struct TConvolutionList {
	std::vector<TConvolutionInfo> convolution_list;

	/*
	 *	void TConvolutionList::Dump()
	 *		- prints convolution info dumps to stdout
	 */
	void Dump();

	/*
	 *	bool TConvolutionList::Create(int n_classifier_num, const TWeakClassifier *p_classifier)
	 *		- creates list of convolutions necessary to evaluate given classifiers
	 *		- n_classifier_num is number of classifiers, p_classifier is list of classifiers
	 *		- returns true on success, false on failure
	 */
	bool Create(int n_classifier_num, const TWeakClassifier *p_classifier);
};

struct TShaderParameters {
	enum {
		conv_PackShader,
		conv_ConvShader
	};

	int n_shader_type;
	int n_offset_x, n_offset_y; // x and y offset of texture
	int n_mip_level; // mip-map texture level
	int n_samples_x, n_samples_y; // number of horiz and vert samples

	float f_pixel_w, f_pixel_h; // pixel size
	float f_min_s, f_min_t, f_max_s, f_max_t; // min and max texcoords

	int n_source_w, n_source_h;
	int n_result_w, n_result_h;

	CGLRenderBuffer *p_render_buffer;
	CGLTexture_2D *p_image;

	TShaderParameters(CGLState *p_state, const TConvolutionInfo &r_t_info, int n_src_w, int n_src_h);

	TShaderParameters(const TShaderParameters &r_t_sp);

	TShaderParameters operator =(const TShaderParameters &r_t_sp);

	void Dump();
};

class CAdHocShaderInfo : public TShaderInfo {
	TShaderInfo::TShader m_t_high;

public:
	CAdHocShaderInfo(const char *p_s_shader_name, bool b_fragment_shader,
		const char *p_s_source_code, const char *p_s_sampler_names = 0);
	~CAdHocShaderInfo();
};

class CConvolutionManager {
protected:
	std::vector<CGLRenderBuffer_FBO*> m_render_buffer_list;
	CGLShaderObject *m_p_pack_shader, *m_p_convolve_shader;
	std::vector<TShaderParameters> m_shader_params_list;
	std::vector<CGLTexture_2D*> m_image_list;

	class CFindGLResource {
	protected:
		GLenum m_n_internal_format;
		int m_n_width, m_n_height;

	public:
		CFindGLResource(int n_width, int n_height, GLenum n_internal_format)
			:m_n_width(n_width), m_n_height(n_height), m_n_internal_format(n_internal_format)
		{}

		bool operator ()(const CGLTexture_2D *p_texture) const
		{
			return p_texture->n_Pixel_Format() == m_n_internal_format &&
				p_texture->n_Width() == m_n_width &&
				p_texture->n_Height() == m_n_height;
		}

		bool operator ()(const CGLRenderBuffer *p_fbo) const
		{
			return p_fbo->n_Color_Format() == m_n_internal_format &&
				p_fbo->n_Width() == m_n_width &&
				p_fbo->n_Height() == m_n_height;
		}
	};

public:
	CConvolutionManager();
	~CConvolutionManager();

	bool CreateConvolutionPlan(CGLState *p_state, const TConvolutionList &r_t_conv_list,
		int n_source_width, int n_source_height, bool b_16_bit_sums = true);

	bool Convolve(CGLState *p_state, CGLTexture_2D *p_source_image);

	inline int n_Convolution_Num() const
	{
		return m_image_list.size();
	}

	inline const CGLTexture_2D &r_Convolution_Image(int n_index) const
	{
		return *m_image_list[n_index];
	}

	inline CGLTexture_2D &r_Convolution_Image(int n_index)
	{
		return *m_image_list[n_index];
	}

protected:
	static void Delete_FBO(CGLRenderBuffer_FBO *p_fbo);
	static void Delete_Texture(CGLTexture_2D *p_tex);
};

/*
const char *p_s_shader_copy_image =
	"uniform sampler2D n_texture;\n"
	"uniform vec4 v_block_shift;\n" // pixel-w, 2 * pixel-w, 3 * pixel-w, 0
	"\n"
	"void main()\n"
	"{\n"
	"    vec4 v_sample;\n"
	"    v_sample.x = texture2D(n_texture, gl_TexCoord[0].xy).x;\n"
	"    v_sample.y = texture2D(n_texture, gl_TexCoord[0].xy + v_block_shift.xw).x;\n"
	"    v_sample.z = texture2D(n_texture, gl_TexCoord[0].xy + v_block_shift.yw).x;\n"
	"    v_sample.w = texture2D(n_texture, gl_TexCoord[0].xy + v_block_shift.zw).x;\n" // pack four grey pixels to RGBA pixel
	"    gl_FragColor = v_sample;\n"
	"}\n";

const char *p_s_shader_convolve_image =
	"uniform sampler2D n_texture;\n"
	"uniform vec2 v_pixel;\n" // pixel-w, pixel-h
	"uniform ivec2 v_kernel;\n" // width, height
	"uniform vec4 v_block_shift;\n" // width * pixel-w, 2 * width * pixel-w, 3 * width * pixel-w, 0
	"\n"
	"void main()\n"
	"{\n"
	"    vec4 v_sum = vec4(.0, .0, .0, .0);\n"
	"    vec2 v_texcoord = gl_TexCoord[0].xy;\n"
	"    for(int i = 0; i != v_kernel.x; ++ i) {\n"
	"        v_texcoord.x = gl_TexCoord[0].x;\n"
	"        for(int j = 0; j != v_kernel.y; ++ j) {\n"
    "            vec4 v_sample;\n"
	"            v_sample.x = texture2D(n_texture, v_texcoord).x;\n"
	"            v_sample.y = texture2D(n_texture, v_texcoord + v_block_shift.xw).x;\n"
	"            v_sample.z = texture2D(n_texture, v_texcoord + v_block_shift.yw).x;\n"
	"            v_sample.w = texture2D(n_texture, v_texcoord + v_block_shift.zw).x;\n" // calculate sum of four blocks at once
	"            v_texcoord.x += v_pixel.x;\n"
	"            v_sum += v_sample\n;"
	"        }\n"
	"        v_texcoord.y += v_pixel.y;\n"
	"    }\n"
	"    gl_FragColor = v_sum;\n"
	"}\n";
*/

#endif //__ADABOOST_CONVOLUTION_SCHEDULER_INCLUDED
