/*
								+-----------------------------------+
								|                                   |
								|  ***  Rasterizer calibrator  ***  |
								|                                   |
								|   Copyright   -tHE SWINe- 2007   |
								|                                   |
								|           RasCalib.cpp            |
								|                                   |
								+-----------------------------------+
*/

#include <vector>
#include "../../berLame_src/CallStack.h"
#ifdef WIN32
#include <windows.h> // gl.h needs some typedefs
#endif
#include <GL/gl.h>
#include <GL/glut.h>
#include <GL/glext.h>
#include <math.h>
#include <algorithm>
#include "../../berLame_src/OpenGL20.h"
#include "../../berLame_src/OpenGLState.h"
#include "../../berLame_src/VertexBufferObject.h"
#include "../../berLame_src/Shader2.h"
#include "../../berLame_src/ShaderSys.h"
#include "../../berLame_src/Texture.h"
#include "../../berLame_src/RenderBuffer2.h"
#include "Load_XML.h"
#include "ConvPlan.h"
#include "RasCalib.h"

/*
 *								=== CRasterizerCalibrator ===
 */

#ifdef RASTERIZER_CALIBRATOR_USE_FRAGMENT_PROGRAM
TShaderInfo CRasterizerCalibrator::display_st_shader("display_st", true,
	"!!ARBfp1.0\n"
    "MOV result.color, fragment.texcoord[0];\n"
    "END\n", 0, 0); // ARB program version

#else // RASTERIZER_CALIBRATOR_USE_FRAGMENT_PROGRAM
TShaderInfo CRasterizerCalibrator::display_st_shader("display_st", true,
	"void main()\n"
	"{\n"
	"    gl_FragColor = gl_TexCoord[0];\n"
	"}\n"); // GLSL version
#endif //RASTERIZER_CALIBRATOR_USE_FRAGMENT_PROGRAM

/*
 *	static bool CRasterizerCalibrator::CalibrateRasterizer(CGLState *p_state,
 *		int n_width, int n_height, TTexCoordRect t_in_texcoords,
 *		TTexCoordRect &r_t_out_texcoords, float f_epsilon, int n_max_steps = 100)
 *		- determines correct texcoords that has to be passed to OpenGL so
 *		  n_width per n_height pixels quad has desired texcoords in it's corner pixels
 *		- p_state is OpenGL state guard
 *		- t_in_texcoords are desired texcoords, r_t_out_texcoords are correct texcoords
 *		- f_epsilon is maximal texcoord deviation
 *		- the method iteratively converges to solution, n_max_steps is maximal number of
 *		  render passes before it either finds solution or fails
 *		- returns true on success, false on failure
 */
bool CRasterizerCalibrator::CalibrateRasterizer(CGLState *p_state,
	int n_width, int n_height, TTexCoordRect t_in_texcoords,
	TTexCoordRect &r_t_out_texcoords, float f_epsilon, int n_max_steps)
{
	if(glGetError() != GL_NO_ERROR) {
		fprintf(stderr, "error: OpenGL errors occured before ::CalibrateRasterizer()\n");
		return false;
	}

	CGLRenderBuffer_FBO *p_render_buffer;
	if(!(p_render_buffer = new CGLRenderBuffer_FBO(n_width, n_height, 1, false,
	   GL_RGBA16F_ARB, false, false, 0, false, false, 0))) {
		fprintf(stderr, "error: error not enough memory\n");
		return false;
	}
	if(glGetError() != GL_NO_ERROR || !p_render_buffer->b_Status()) {
		fprintf(stderr, "error: error creating render-buffer\n");
		return false;
	}
	// create render buffer

	CGLShaderObject *p_shader;
	if(!(p_shader = new CGLShaderObject(0, &display_st_shader))) {
		delete p_render_buffer;
		fprintf(stderr, "error: error not enough memory\n");
		return false;
	}
	char *p_s_compile = 0, *p_s_link = 0;
	bool b_result = p_shader->Compile_Link(p_state, p_s_compile, p_s_link);
	if(p_s_compile) {
		fprintf(stderr, "shader compile log:\n%s\n", p_s_compile);
		delete[] p_s_compile;
	}
	if(p_s_link) {
		fprintf(stderr, "shader link log:\n%s\n", p_s_link);
		delete[] p_s_link;
	}
	// create shader

	glViewport(0, 0, n_width, n_height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// setup new viewport and ortho view

	bool b_return_value = false;
	do {
		if(!b_result || !p_shader->b_Status()) {
			fprintf(stderr, "error: error creating shader\n");
			break;
		}
		// create shader to get raster uv's

		p_render_buffer->Bind();
		p_shader->Bind(p_state);
		// bind render buffer, shader

		p_state->DisableDepthTest();
		p_state->DisableAlphaTest();
		p_state->DisableCullFace();
		p_state->DisableBlend();
		// need all those disabaled

		r_t_out_texcoords = t_in_texcoords;
		// start with original texcoords

		for(int i = 0; i < n_max_steps; ++ i) {
			glBegin(GL_QUADS);
			glTexCoord2f(r_t_out_texcoords.f_min_s, r_t_out_texcoords.f_min_t);
			glVertex2f(-1,  1);
			glTexCoord2f(r_t_out_texcoords.f_max_s, r_t_out_texcoords.f_min_t);
			glVertex2f( 1,  1);
			glTexCoord2f(r_t_out_texcoords.f_max_s, r_t_out_texcoords.f_max_t);
			glVertex2f( 1, -1);
			glTexCoord2f(r_t_out_texcoords.f_min_s, r_t_out_texcoords.f_max_t);
			glVertex2f(-1, -1);
			glEnd();
			// draw fullscreen quad

			float p_coords[8];
			glReadPixels(0, 0, 1, 1, GL_RGBA, GL_FLOAT, p_coords);
			glReadPixels(p_render_buffer->n_Width() - 1, p_render_buffer->n_Height() - 1,
				1, 1, GL_RGBA, GL_FLOAT, p_coords + 4);
			// get corner coordinates

			if(b_zero(t_in_texcoords.f_min_s - p_coords[0], f_epsilon) &&
			   b_zero(t_in_texcoords.f_max_t - p_coords[1], f_epsilon) &&
			   b_zero(t_in_texcoords.f_max_s - p_coords[4], f_epsilon) &&
			   b_zero(t_in_texcoords.f_min_t - p_coords[5], f_epsilon)) {
				b_return_value = true;
				break;
			}
			// compare texcoords

			const float f_weight = .75f;
			r_t_out_texcoords.f_min_s += (t_in_texcoords.f_min_s - p_coords[0]) * f_weight;
			r_t_out_texcoords.f_max_t += (t_in_texcoords.f_max_t - p_coords[1]) * f_weight;
			r_t_out_texcoords.f_max_s += (t_in_texcoords.f_max_s - p_coords[4]) * f_weight;
			r_t_out_texcoords.f_min_t += (t_in_texcoords.f_min_t - p_coords[5]) * f_weight;
			// offset texcoords
		}
		// iteratively aproximate solution

		if(p_shader->n_Api() == shader_api_Core_shader)
			p_state->BindProgramObject_GL2(0);
		else if(p_shader->n_Api() == shader_api_ARB_shader)
			p_state->BindProgramObject(0);
		else /*if(p_shader->n_Api() == shader_api_ARB_program)*/
			p_state->DisableFragmentProgram();
		// release shader

		p_render_buffer->Release();
		// release render buffer
	} while(0);

	delete p_shader;
	delete p_render_buffer;

	if(glGetError() != GL_NO_ERROR) {
		fprintf(stderr, "error: OpenGL errors occured\n");
		return false;
	}

	return b_return_value;
}


/*
 *	static float CRasterizerCalibrator::f_fabs(float f_x)
 *		- returns absolute value of f_x
 *		- want to avoid conversion to double and back again
 *		- note this is most likely not going to compile to effective code on GCC
 *
 *		- not used anymore
 */
float CRasterizerCalibrator::f_fabs(float f_x)
{
	union {
		float f;
		unsigned char c[8]; // required by ANSI
	} t_fi = {f_x};
	t_fi.c[3] &= 0x7f;
	return t_fi.f;
}

/*
 *	static inline bool CRasterizerCalibrator::b_zero(float f_x, float f_epsilon)
 *		- returns true in case f_x is zero, with tolerance f_epsilon;
 *		  otherwise returns false
 */
bool CRasterizerCalibrator::b_zero(float f_x, float f_epsilon)
{
	return f_x > -f_epsilon && f_x < f_epsilon;
}

/*
 *								=== ~CRasterizerCalibrator ===
 */
